import cupy as cp
import numpy as np
import random


__outward_functions = [
    "diagonal_targ",
    "diagonal_targs",
    "normal_targ",
    "normal_targs",
    "control_targ",
    "diagonal_ctargs",
    "control_ctargs",
    "normal_ctargs",
    "ctrl_normal_targs",
    "normal_normal_targs",
    "diagonal_normal_targs",
    "swap_targ",
    "reverse_targ",
    "reverse_ctargs",
    "swap_targs",
    "reverse_more",
    "diagonal_more",
    "swap_tmore",
    "measured_prob_calculate",
    "apply_measuregate",
    "apply_resetgate"
]


Diagonal_Multiply_targ_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Diagonal2x2Multiply(int parg, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        vec[_0] = vec[_0]*mat[0];
        vec[_1] = vec[_1]*mat[3];
    }
    ''', 'Diagonal2x2Multiply')


Diagonal_Multiply_targ_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Diagonal2x2Multiply(int parg, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        vec[_0] = vec[_0]*mat[0];
        vec[_1] = vec[_1]*mat[3];
    }
    ''', 'Diagonal2x2Multiply')


Diagonal_Multiply_targs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Diagonal4x4Multiply(int high, int low, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _1 + offset2;

        vec[_0] = vec[_0]*mat[0];
        vec[_1] = vec[_1]*mat[5];
        vec[_2] = vec[_2]*mat[10];
        vec[_3] = vec[_3]*mat[15];
    }
    ''', 'Diagonal4x4Multiply')


Diagonal_Multiply_targs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Diagonal4x4Multiply(int high, int low, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _1 + offset2;

        vec[_0] = vec[_0]*mat[0];
        vec[_1] = vec[_1]*mat[5];
        vec[_2] = vec[_2]*mat[10];
        vec[_3] = vec[_3]*mat[15];
    }
    ''', 'Diagonal4x4Multiply')


Based_InnerProduct_targ_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Based2x2InnerProduct(int parg, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[0] + vec[_1]*mat[1];
        vec[_1] = temp_0*mat[2] + vec[_1]*mat[3];
    }
    ''', 'Based2x2InnerProduct')


Based_InnerProduct_targ_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Based2x2InnerProduct(int parg, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[0] + vec[_1]*mat[1];
        vec[_1] = temp_0*mat[2] + vec[_1]*mat[3];
    }
    ''', 'Based2x2InnerProduct')


Based_InnerProduct_targs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Based4x4InnerProduct(int t0, int t1, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << t0;
        const int offset2 = 1 << t1;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw=0, _0=0;

        if (t0 > t1){
            gw = label >> t1 << (t1 + 1);
            _0 = (gw >> t0 << (t0 + 1)) + (gw & (offset1 - offset2)) + (label & mask2);
        }
        else{
            gw = label >> t0 << (t0 + 1);
            _0 = (gw >> t1 << (t1 + 1)) + (gw & (offset2 - offset1)) + (label & mask1);
        }

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _1 + offset2;

        complex<float> temp_0 = vec[_0], temp_1 = vec[_1], temp_2 = vec[_2];
        vec[_0] = vec[_0]*mat[0] + vec[_1]*mat[1] + vec[_2]*mat[2] + vec[_3]*mat[3];
        vec[_1] = temp_0*mat[4] + vec[_1]*mat[5] + vec[_2]*mat[6] + vec[_3]*mat[7];
        vec[_2] = temp_0*mat[8] + temp_1*mat[9] + vec[_2]*mat[10] + vec[_3]*mat[11];
        vec[_3] = temp_0*mat[12] + temp_1*mat[13] + temp_2*mat[14] + vec[_3]*mat[15];
    }
    ''', 'Based4x4InnerProduct')


Based_InnerProduct_targs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Based4x4InnerProduct(int t0, int t1, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << t0;
        const int offset2 = 1 << t1;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw=0, _0=0;

        if (t0 > t1){
            gw = label >> t1 << (t1 + 1);
            _0 = (gw >> t0 << (t0 + 1)) + (gw & (offset1 - offset2)) + (label & mask2);
        }
        else{
            gw = label >> t0 << (t0 + 1);
            _0 = (gw >> t1 << (t1 + 1)) + (gw & (offset2 - offset1)) + (label & mask1);
        }

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _1 + offset2;

        complex<double> temp_0 = vec[_0], temp_1 = vec[_1], temp_2 = vec[_2];
        vec[_0] = vec[_0]*mat[0] + vec[_1]*mat[1] + vec[_2]*mat[2] + vec[_3]*mat[3];
        vec[_1] = temp_0*mat[4] + vec[_1]*mat[5] + vec[_2]*mat[6] + vec[_3]*mat[7];
        vec[_2] = temp_0*mat[8] + temp_1*mat[9] + vec[_2]*mat[10] + vec[_3]*mat[11];
        vec[_3] = temp_0*mat[12] + temp_1*mat[13] + temp_2*mat[14] + vec[_3]*mat[15];
    }
    ''', 'Based4x4InnerProduct')


Controlled_Multiply_targ_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled2x2Multiply(int parg, const complex<float> val, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _1 = (label >> parg << (parg + 1)) + offset + (label & (offset - 1));

        vec[_1] = vec[_1]*val;
    }
    ''', 'Controlled2x2Multiply')


Controlled_Multiply_targ_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled2x2Multiply(int parg, const complex<double> val, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _1 = (label >> parg << (parg + 1)) + offset + (label & (offset - 1));

        vec[_1] = vec[_1]*val;
    }
    ''', 'Controlled2x2Multiply')


Controlled_Multiply_ctargs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4Multiply(const complex<float>* mat, complex<float>* vec, int c_index, int t_index) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        int _1 = _0 + offset_t;

        vec[_0] = vec[_0]*mat[10];
        vec[_1] = vec[_1]*mat[15];
    }
    ''', 'Controlled4x4Multiply')


Controlled_Multiply_ctargs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4Multiply(const complex<double>* mat, complex<double>* vec, int c_index, int t_index){
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        int _1 = _0 + offset_t;

        vec[_0] = vec[_0]*mat[10];
        vec[_1] = vec[_1]*mat[15];
    }
    ''', 'Controlled4x4Multiply')


Controlled_Product_ctargs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4Product(const complex<float> val, complex<float>* vec, int c_index, int t_index) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        _0 = _0 + offset_t;
        vec[_0] = vec[_0]*val;
    }
    ''', 'Controlled4x4Product')


Controlled_Product_ctargs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4Product(const complex<double> val, complex<double>* vec, int c_index, int t_index){
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        _0 = _0 + offset_t;
        vec[_0] = vec[_0]*val;
    }
    ''', 'Controlled4x4Product')


Controlled_InnerProduct_ctargs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4InnerProduct(const complex<float>* mat, complex<float>* vec, int c_index, int t_index) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        int _1 = _0 + offset_t;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[10] + vec[_1]*mat[11];
        vec[_1] = temp_0*mat[14] + vec[_1]*mat[15];
    }
    ''', 'Controlled4x4InnerProduct')


Controlled_InnerProduct_ctargs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4InnerProduct(const complex<double>* mat, complex<double>* vec, int c_index, int t_index) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        int _1 = _0 + offset_t;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[10] + vec[_1]*mat[11];
        vec[_1] = temp_0*mat[14] + vec[_1]*mat[15];
    }
    ''', 'Controlled4x4InnerProduct')


Controlled_MultiplySwap_ctargs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4MultiSwap(const complex<float>* mat, complex<float>* vec, int c_index, int t_index) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        int _1 = _0 + offset_t;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_1]*mat[11];
        vec[_1] = temp_0*mat[14];
    }
    ''', 'Controlled4x4MultiSwap')


Controlled_MultiplySwap_ctargs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4MultiSwap(const complex<double>* mat, complex<double>* vec, int c_index, int t_index) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c = 1 << c_index;
        const int offset_t = 1 << t_index;
        const int mask_c = offset_c - 1;
        const int mask_t = offset_t - 1;

        int gw=0, _0=0;

        if (t_index > c_index){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c)) + (label & mask_c);
        }
        else
        {
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t)) + (label & mask_t);
        }

        int _1 = _0 + offset_t;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_1]*mat[11];
        vec[_1] = temp_0*mat[14];
    }
    ''', 'Controlled4x4MultiSwap')


Controlled_Swap_targs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4Swap(int high, int low, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;

        complex<float> temp_0 = vec[_1];
        vec[_1] = vec[_2]*mat[6];
        vec[_2] = temp_0*mat[9];
    }
    ''', 'Controlled4x4Swap')


Controlled_Swap_targs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled4x4Swap(int high, int low, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;

        complex<double> temp_0 = vec[_1];
        vec[_1] = vec[_2]*mat[6];
        vec[_2] = temp_0*mat[9];
    }
    ''', 'Controlled4x4Swap')


Completed_MxIP_targs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void CompletedMxIP(int high, int low, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _2 + offset1;

        vec[_0] = vec[_0]*mat[0];
        vec[_3] = vec[_3]*mat[15];

        complex<float> temp_0 = vec[_1];
        vec[_1] = vec[_1]*mat[5] + vec[_2]*mat[6];
        vec[_2] = temp_0*mat[9] + vec[_2]*mat[10];
    }
    ''', 'CompletedMxIP')


Completed_MxIP_targs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void CompletedMxIP(int high, int low, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _2 + offset1;

        vec[_0] = vec[_0]*mat[0];
        vec[_3] = vec[_3]*mat[15];

        complex<double> temp_0 = vec[_1];
        vec[_1] = vec[_1]*mat[5] + vec[_2]*mat[6];
        vec[_2] = temp_0*mat[9] + vec[_2]*mat[10];
    }
    ''', 'CompletedMxIP')


Completed_IPxIP_targs_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void CompletedIPxIP(int high, int low, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _2 + offset1;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[0] + vec[_3]*mat[3];
        vec[_3] = temp_0*mat[12] + vec[_3]*mat[15];

        complex<float> temp_1 = vec[_1];
        vec[_1] = vec[_1]*mat[5] + vec[_2]*mat[6];
        vec[_2] = temp_1*mat[9] + vec[_2]*mat[10];
    }
    ''', 'CompletedIPxIP')


Completed_IPxIP_targs_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void CompletedIPxIP(int high, int low, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << low;
        const int offset2 = 1 << high;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw = label >> low << (low + 1);
        int _0 = (gw >> high << (high + 1)) + (gw & (offset2 - offset1)) + (label & mask1);

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _2 + offset1;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[0] + vec[_3]*mat[3];
        vec[_3] = temp_0*mat[12] + vec[_3]*mat[15];

        complex<double> temp_1 = vec[_1];
        vec[_1] = vec[_1]*mat[5] + vec[_2]*mat[6];
        vec[_2] = temp_1*mat[9] + vec[_2]*mat[10];
    }
    ''', 'CompletedIPxIP')


Diagonal_Multiply_normal_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void DiagxNormal(int t0, int t1, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset1 = 1 << t0;
        const int offset2 = 1 << t1;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw=0, _0=0;

        if (t0 > t1){
            gw = label >> t1 << (t1 + 1);
            _0 = (gw >> t0 << (t0 + 1)) + (gw & (offset1 - offset2)) + (label & mask2);
        }
        else{
            gw = label >> t0 << (t0 + 1);
            _0 = (gw >> t1 << (t1 + 1)) + (gw & (offset2 - offset1)) + (label & mask1);
        }
        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _1 + offset2;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[0] + vec[_1]*mat[1];
        vec[_1] = temp_0*mat[4] + vec[_1]*mat[5];
        complex<float> temp_2 = vec[_2];
        vec[_2] = vec[_2]*mat[10] + vec[_3]*mat[11];
        vec[_3] =temp_2*mat[14] + vec[_3]*mat[15];
    }
    ''', 'DiagxNormal')


Diagonal_Multiply_normal_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void DiagxNormal(int t0, int t1, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        const int offset1 = 1 << t0;
        const int offset2 = 1 << t1;
        const int mask1 = offset1 - 1;
        const int mask2 = offset2 - 1;

        int gw=0, _0=0;

        if (t0 > t1){
            gw = label >> t1 << (t1 + 1);
            _0 = (gw >> t0 << (t0 + 1)) + (gw & (offset1 - offset2)) + (label & mask2);
        }
        else{
            gw = label >> t0 << (t0 + 1);
            _0 = (gw >> t1 << (t1 + 1)) + (gw & (offset2 - offset1)) + (label & mask1);
        }

        int _1 = _0 + offset1;
        int _2 = _0 + offset2;
        int _3 = _1 + offset2;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_0]*mat[0] + vec[_1]*mat[1];
        vec[_1] = temp_0*mat[4] + vec[_1]*mat[5];

        complex<double> temp_2 = vec[_2];
        vec[_2] = vec[_2]*mat[10] + vec[_3]*mat[11];
        vec[_3] =temp_2*mat[14] + vec[_3]*mat[15];
    }
    ''', 'DiagxNormal')


RDiagonal_Swap_targ_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void RDiag2x2Swap(int parg, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_1];
        vec[_1] = temp_0;
    }
    ''', 'RDiag2x2Swap')


RDiagonal_Swap_targ_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void RDiag2x2Swap(int parg, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_1];
        vec[_1] = temp_0;
    }
    ''', 'RDiag2x2Swap')


RDiagonal_MultiplySwap_targ_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void RDiag2x2MultiSwap(int parg, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_1]*mat[1];
        vec[_1] = temp_0*mat[2];
    }
    ''', 'RDiag2x2MultiSwap')


RDiagonal_MultiplySwap_targ_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void RDiag2x2MultiSwap(int parg, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        int offset = 1 << parg;

        int _0 = (label >> parg << (parg + 1)) + (label & (offset - 1));
        int _1 = _0 +  offset;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_1]*mat[1];
        vec[_1] = temp_0*mat[2];
    }
    ''', 'RDiag2x2MultiSwap')


Controlled_Swap_more_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled8x8Swap(int high, int low, int t_index, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c1 = 1 << low;
        const int offset_c2 = 1 << high;
        const int offset_t = 1 << t_index;
        const int maskc1 = offset_c1 - 1;
        const int maskc2 = offset_c2 - 1;
        const int mask_t = offset_t - 1;

        int gw = 0, _0 = 0;

        if (t_index < low){
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c1 + (gw >> low << (low + 1)) + (gw & (offset_c1 - offset_t)) + (label & mask_t);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & maskc2);
        }else if(t_index < high){
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c1)) + (label & maskc1);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & maskc2);
        }else{
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> high << (high + 1)) + (gw & (offset_c2 - offset_c1)) + (label & maskc1);
            _0 = offset_c2 + (_0 >> t_index << (t_index + 1)) + (_0 & mask_t);
        }

        int _1 = _0 + offset_t;

        complex<float> temp_0 = vec[_0];
        vec[_0] = vec[_1];
        vec[_1] = temp_0;
    }
    ''', 'Controlled8x8Swap')


Controlled_Swap_more_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled8x8Swap(int high, int low, int t_index, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c1 = 1 << low;
        const int offset_c2 = 1 << high;
        const int offset_t = 1 << t_index;
        const int maskc1 = offset_c1 - 1;
        const int maskc2 = offset_c2 - 1;
        const int mask_t = offset_t - 1;

        int gw = 0, _0 = 0;

        if (t_index < low){
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c1 + (gw >> low << (low + 1)) + (gw & (offset_c1 - offset_t)) + (label & mask_t);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & maskc2);
        }else if(t_index < high){
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c1)) + (label & maskc1);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & maskc2);
        }else{
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> high << (high + 1)) + (gw & (offset_c2 - offset_c1)) + (label & maskc1);
            _0 = offset_c2 + (_0 >> t_index << (t_index + 1)) + (_0 & mask_t);
        }

        int _1 = _0 + offset_t;

        complex<double> temp_0 = vec[_0];
        vec[_0] = vec[_1];
        vec[_1] = temp_0;
    }
    ''', 'Controlled8x8Swap')


Controlled_Multiply_more_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled8x8Multiply(int high, int low, int t_index, const complex<float>* mat, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c1 = 1 << low;
        const int offset_c2 = 1 << high;
        const int offset_t = 1 << t_index;
        const int mask1 = offset_c1 - 1;
        const int mask2 = offset_c2 - 1;
        const int mask_t = offset_t - 1;

        int gw = 0, _0 = 0;

        if (t_index < low){
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c1 + (gw >> low << (low + 1)) + (gw & (offset_c1 - offset_t)) + (label & mask_t);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & mask2);
        }else if(t_index < high){
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c1)) + (label & mask1);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & mask2);
        }else{
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> high << (high + 1)) + (gw & (offset_c2 - offset_c1)) + (label & mask1);
            _0 = offset_c2 + (_0 >> t_index << (t_index + 1)) + (_0 & mask_t);
        }

        int _1 = _0 + offset_t;

        vec[_0] = vec[_0]*mat[54];
        vec[_1] = vec[_1]*mat[63];
    }
    ''', 'Controlled8x8Multiply')


Controlled_Multiply_more_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled8x8Multiply(int high, int low, int t_index, const complex<double>* mat, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_c1 = 1 << low;
        const int offset_c2 = 1 << high;
        const int offset_t = 1 << t_index;
        const int mask1 = offset_c1 - 1;
        const int mask2 = offset_c2 - 1;
        const int mask_t = offset_t - 1;

        int gw = 0, _0 = 0;

        if (t_index < low){
            gw = label >> t_index << (t_index + 1);
            _0 = offset_c1 + (gw >> low << (low + 1)) + (gw & (offset_c1 - offset_t)) + (label & mask_t);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & mask2);
        }else if(t_index < high){
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> t_index << (t_index + 1)) + (gw & (offset_t - offset_c1)) + (label & mask1);
            _0 = offset_c2 + (_0 >> high << (high + 1)) + (_0 & mask2);
        }else{
            gw = label >> low << (low + 1);
            _0 = offset_c1 + (gw >> high << (high + 1)) + (gw & (offset_c2 - offset_c1)) + (label & mask1);
            _0 = offset_c2 + (_0 >> t_index << (t_index + 1)) + (_0 & mask_t);
        }

        int _1 = _0 + offset_t;

        vec[_0] = vec[_0]*mat[54];
        vec[_1] = vec[_1]*mat[63];
    }
    ''', 'Controlled8x8Multiply')


Controlled_Swap_tmore_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled8x8Swapt(int high, int low, int c_index, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_t1 = 1 << low;
        const int offset_t2 = 1 << high;
        const int offset_c = 1 << c_index;
        const int mask1 = offset_t1 - 1;
        const int mask2 = offset_t2 - 1;
        const int mask_c = offset_c - 1;

        int gw = 0, _0 = 0;

        if (c_index < low){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> low << (low + 1)) + (gw & (offset_t1 - offset_c)) + (label & mask_c);
            _0 = (_0 >> high << (high + 1)) + (_0 & mask2);
        }else if(c_index < high){
            gw = label >> low << (low + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t1)) + (label & mask1);
            _0 = (_0 >> high << (high + 1)) + (_0 & mask2);
        }else{
            gw = label >> low << (low + 1);
            _0 = (gw >> high << (high + 1)) + (gw & (offset_t2 - offset_t1)) + (label & mask1);
            _0 = offset_c + (_0 >> c_index << (c_index + 1)) + (_0 & mask_c);
        }

        int _1 = _0 + offset_t1;
        int _2 = _0 + offset_t2;

        complex<float> temp_0 = vec[_1];
        vec[_1] = vec[_2];
        vec[_2] = temp_0;
    }
    ''', 'Controlled8x8Swapt')


Controlled_Swap_tmore_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void Controlled8x8Swapt(int high, int low, int c_index, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        const int offset_t1 = 1 << low;
        const int offset_t2 = 1 << high;
        const int offset_c = 1 << c_index;
        const int mask1 = offset_t1 - 1;
        const int mask2 = offset_t2 - 1;
        const int mask_c = offset_c - 1;

        int gw = 0, _0 = 0;

        if (c_index < low){
            gw = label >> c_index << (c_index + 1);
            _0 = offset_c + (gw >> low << (low + 1)) + (gw & (offset_t1 - offset_c)) + (label & mask_c);
            _0 = (_0 >> high << (high + 1)) + (_0 & mask2);
        }else if(c_index < high){
            gw = label >> low << (low + 1);
            _0 = offset_c + (gw >> c_index << (c_index + 1)) + (gw & (offset_c - offset_t1)) + (label & mask1);
            _0 = (_0 >> high << (high + 1)) + (_0 & mask2);
        }else{
            gw = label >> low << (low + 1);
            _0 = (gw >> high << (high + 1)) + (gw & (offset_t2 - offset_t1)) + (label & mask1);
            _0 = offset_c + (_0 >> c_index << (c_index + 1)) + (_0 & mask_c);
        }

        int _1 = _0 + offset_t1;
        int _2 = _0 + offset_t2;

        complex<double> temp_0 = vec[_1];
        vec[_1] = vec[_2];
        vec[_2] = temp_0;
    }
    ''', 'Controlled8x8Swapt')


def diagonal_targ(t_index, mat, vec, vec_bit, sync: bool = False):
    """
    Diagonal matrix (2x2) dot vector
        [[a, 0],    *   vec
         [0, b]]
    """
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Diagonal_Multiply_targ_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, mat, vec)
        )
    else:
        Diagonal_Multiply_targ_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def diagonal_targs(t_indexes, mat, vec, vec_bit, sync: bool = False):
    """
    Diagonal matrix (4x4) dot vector
        [[a, 0, 0, 0],    *   vec
         [0, b, 0, 0],
         [0, 0, c, 0],
         [0, 0, 0, d]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if t_indexes[0] > t_indexes[1]:
        high, low = t_indexes[0], t_indexes[1]
    else:
        high, low = t_indexes[1], t_indexes[0]

    if vec.dtype == np.complex64:
        Diagonal_Multiply_targs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )
    else:
        Diagonal_Multiply_targs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def normal_targ(t_index, mat, vec, vec_bit, sync: bool = False):
    """
    Based matrix (2x2) dot vector
        [[a, b],    *   vec
         [c, d]]
    """
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Based_InnerProduct_targ_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, mat, vec)
        )
    else:
        Based_InnerProduct_targ_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def normal_targs(t_indexes, mat, vec, vec_bit, sync: bool = False):
    """
    Based matrix (4x4) dot vector
        [[a, b, c, d],    *   vec
         [e, f, g, h],
         [i, j, k, l],
         [m, n, o, p]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Based_InnerProduct_targs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_indexes[0], t_indexes[1], mat, vec)
        )
    else:
        Based_InnerProduct_targs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_indexes[0], t_indexes[1], mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def control_targ(t_index, val, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (2x2) dot vector
        [[1, 0],    *   vec
         [0, a]]
    """
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Controlled_Multiply_targ_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, val, vec)
        )
    else:
        Controlled_Multiply_targ_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, val, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def diagonal_ctargs(c_index, t_index, mat, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (4x4) dot vector
     e.g.   [[1, 0, 0, 0],    *   vec
             [0, 1, 0, 0],
             [0, 0, a, 0],
             [0, 0, 0, b]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Controlled_Multiply_ctargs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (mat, vec, c_index, t_index)
        )
    else:
        Controlled_Multiply_ctargs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (mat, vec, c_index, t_index)
        )

    if sync:
        cp.cuda.Device().synchronize()


def control_ctargs(c_index, t_index, value, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (4x4) dot vector
     e.g.   [[1, 0, 0, 0],    *   vec
             [0, 1, 0, 0],
             [0, 0, 1, 0],
             [0, 0, 0, a]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Controlled_Product_ctargs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (value, vec, c_index, t_index)
        )
    else:
        Controlled_Product_ctargs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (value, vec, c_index, t_index)
        )

    if sync:
        cp.cuda.Device().synchronize()


def normal_ctargs(c_index, t_index, mat, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (4x4) dot vector
     e.g.   [[1, 0, 0, 0],    *   vec
             [0, 1, 0, 0],
             [0, 0, a, b],
             [0, 0, c, d]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Controlled_InnerProduct_ctargs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (mat, vec, c_index, t_index)
        )
    else:
        Controlled_InnerProduct_ctargs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (mat, vec, c_index, t_index)
        )

    if sync:
        cp.cuda.Device().synchronize()


def ctrl_normal_targs(t_indexes, mat, vec, vec_bit, sync: bool = False):
    """
    Completed matrix (4x4) dot vector
            [[A, 0, 0, 0],    *   vec
             [0, c, d, 0],
             [0, e, f, 0],
             [0, 0, 0, B]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if t_indexes[0] > t_indexes[1]:
        high, low = t_indexes[0], t_indexes[1]
    else:
        high, low = t_indexes[1], t_indexes[0]

    if vec.dtype == np.complex64:
        Completed_MxIP_targs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )
    else:
        Completed_MxIP_targs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def normal_normal_targs(t_indexes, mat, vec, vec_bit, sync: bool = False):
    """
    Completed matrix (4x4) dot vector
            [[A, 0, 0, B],    *   vec
             [0, a, b, 0],
             [0, c, d, 0],
             [C, 0, 0, D]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if t_indexes[0] > t_indexes[1]:
        high, low = t_indexes[0], t_indexes[1]
    else:
        high, low = t_indexes[1], t_indexes[0]

    if vec.dtype == np.complex64:
        Completed_IPxIP_targs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )
    else:
        Completed_IPxIP_targs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def diagonal_normal_targs(t_indexes, mat, vec, vec_bit, sync: bool = False):
    """
    Completed matrix (4x4) dot vector
            [[A, B, 0, 0],    *   vec
             [C, D, 0, 0],
             [0, 0, a, b],
             [0, 0, c, d]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Diagonal_Multiply_normal_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_indexes[0], t_indexes[1], mat, vec)
        )
    else:
        Diagonal_Multiply_normal_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_indexes[0], t_indexes[1], mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def swap_targ(t_index, vec, vec_bit, sync: bool = False):
    """
    reverse diagonal matrix (2x2) dot vector
        [[0, 1],        *       vec
         [1, 0]]
    """
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        RDiagonal_Swap_targ_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, vec)
        )
    else:
        RDiagonal_Swap_targ_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def reverse_targ(t_index, mat, vec, vec_bit, sync: bool = False):
    """
    reverse diagonal matrix (2x2) dot vector
        [[0, a],        *       vec
         [b, 0]]
    """
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        RDiagonal_MultiplySwap_targ_single_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, mat, vec)
        )
    else:
        RDiagonal_MultiplySwap_targ_double_kernel(
            (block_num,),
            (thread_per_block,),
            (t_index, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def reverse_ctargs(c_index, t_index, mat, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (4x4) dot vector
     e.g.   [[1, 0, 0, 0],    *   vec
             [0, 1, 0, 0],
             [0, 0, 0, a],
             [0, 0, b, 0]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        Controlled_MultiplySwap_ctargs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (mat, vec, c_index, t_index)
        )
    else:
        Controlled_MultiplySwap_ctargs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (mat, vec, c_index, t_index)
        )

    if sync:
        cp.cuda.Device().synchronize()


def swap_targs(t_indexes, mat, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (4x4) dot vector
            [[1, 0, 0, 0],    *   vec
             [0, 0, a, 0],
             [0, b, 0, 0],
             [0, 0, 0, 1]]
    """
    task_number = 1 << (vec_bit - 2)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if t_indexes[0] > t_indexes[1]:
        high, low = t_indexes[0], t_indexes[1]
    else:
        high, low = t_indexes[1], t_indexes[0]

    if vec.dtype == np.complex64:
        Controlled_Swap_targs_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )
    else:
        Controlled_Swap_targs_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def reverse_more(c_indexes, t_index, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (8x8) dot vector
       [[1, 0, 0, 0, 0, 0, 0, 0],       *       vec
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 1, 0]]
    """
    task_number = 1 << (vec_bit - 3)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if c_indexes[0] > c_indexes[1]:
        high, low = c_indexes[0], c_indexes[1]
    else:
        high, low = c_indexes[1], c_indexes[0]

    if vec.dtype == np.complex64:
        Controlled_Swap_more_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, t_index, vec)
        )
    else:
        Controlled_Swap_more_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, t_index, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def diagonal_more(c_indexes, t_index, mat, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (8x8) dot vector
       [[1, 0, 0, 0, 0, 0, 0, 0],       *       vec
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, a, 0],
        [0, 0, 0, 0, 0, 0, 0, b]]
    """
    task_number = 1 << (vec_bit - 3)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if c_indexes[0] > c_indexes[1]:
        high, low = c_indexes[0], c_indexes[1]
    else:
        high, low = c_indexes[1], c_indexes[0]

    if vec.dtype == np.complex64:
        Controlled_Multiply_more_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, t_index, mat, vec)
        )
    else:
        Controlled_Multiply_more_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, t_index, mat, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def swap_tmore(t_indexes, c_index, vec, vec_bit, sync: bool = False):
    """
    Controlled matrix (8x8) dot vector
       [[1, 0, 0, 0, 0, 0, 0, 0],       *       vec
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1]]
    """
    task_number = 1 << (vec_bit - 3)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if t_indexes[0] > t_indexes[1]:
        high, low = t_indexes[0], t_indexes[1]
    else:
        high, low = t_indexes[1], t_indexes[0]

    if vec.dtype == np.complex64:
        Controlled_Swap_tmore_single_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, c_index, vec)
        )
    else:
        Controlled_Swap_tmore_double_kernel(
            (block_num,),
            (thread_per_block,),
            (high, low, c_index, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


"""
Special Gates: MeasureGate, ResetGate and PermGate
"""
prop_add_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void ProbAddSingle(const int index, complex<float>* vec, complex<float>* out) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        out[label] = abs(vec[_0]) * abs(vec[_0]);
    }
    ''', 'ProbAddSingle')


prop_add_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void ProbAddDouble(const int index, complex<double>* vec, complex<double>* out) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        out[label] = abs(vec[_0]) * abs(vec[_0]);
    }
    ''', 'ProbAddDouble')


MeasureGate_prop = cp.ReductionKernel(
    'T x',
    'T y',
    'x',
    'a + b',
    'y = abs(a)',
    '0',
    'MeasureGate_prop'
)


mn_measureprob_calculator = cp.ReductionKernel(
    'T x',
    'T y',
    'x',
    'a + b',
    'y = abs(a)*abs(a)',
    '0',
    'mn_measureprob_calculator'
)


MeasureGate0_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void MeasureGate0Single(const int index, const float generation, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        vec[_0] = vec[_0] * generation;
        vec[_0 + (1 << index)] = complex<float>(0, 0);
    }
    ''', 'MeasureGate0Single')


MeasureGate1_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void MeasureGate1Single(const int index, const float generation, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        int _1 = _0 + (1 << index);
        vec[_0] = complex<float>(0, 0);
        vec[_1] = vec[_1] * generation;
    }
    ''', 'MeasureGate1Single')


MeasureGate0_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void MeasureGate0Double(const int index, const double generation, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        vec[_0] = vec[_0] * generation;
        vec[_0 + (1 << index)] = complex<double>(0, 0);
    }
    ''', 'MeasureGate0Double')


MeasureGate1_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void MeasureGate1Double(const int index, const double generation, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        int _1 = _0 + (1 << index);
        vec[_0] = complex<double>(0, 0);
        vec[_1] = vec[_1] * generation;
    }
    ''', 'MeasureGate1Double')


ResetGate0_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void ResetGate0Float(const int index, const float generation, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        vec[_0] = vec[_0] / generation;
        vec[_0 + (1 << index)] = complex<float>(0, 0);
    }
    ''', 'ResetGate0Float')


ResetGate1_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void ResetGate1Float(const int index, const float generation, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        int _1 = _0 + (1 << index);

        vec[_0] = vec[_1];
        vec[_1] = complex<float>(0, 0);
    }
    ''', 'ResetGate1Float')


ResetGate0_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void ResetGate0Double(const int index, const double generation, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _0 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1));
        vec[_0] = vec[_0] / generation;
        vec[_0 + (1 << index)] = complex<double>(0, 0);
    }
    ''', 'ResetGate0Double')


ResetGate1_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void ResetGate1Double(const int index, const double generation, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;
        int _1 = (label & ((1 << index) - 1))
                + (label >> index << (index + 1))
                + (1 << index);

        vec[label] = vec[_1];
    }
    ''', 'ResetGate1Double')


def measured_prob_calculate(index, vec, vec_bit, all_measured: bool = False, sync: bool = False):
    # Deal with the whole vector state measured, only happen for multi-nodes simulator
    if all_measured:
        prob = mn_measureprob_calculator(vec)
        return prob.real

    # Kernel function preparation
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block
    kernel_functions = prop_add_double_kernel if vec.dtype == np.complex128 else prop_add_single_kernel

    # Calculated the probability of measured 1 at current index
    out = cp.empty(task_number, dtype=vec.dtype)
    kernel_functions(
        (block_num, ),
        (thread_per_block, ),
        (index, vec, out)
    )

    prob = MeasureGate_prop(out, axis=0).real

    if sync:
        cp.cuda.Device().synchronize()

    return prob


def apply_measuregate(index, vec, vec_bit, prob, sync: bool = False):
    """
    Measure Gate Measure.
    """
    # Kernel function preparation
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block
    if vec.dtype == np.complex64:
        kernel_functions = (MeasureGate0_single_kernel, MeasureGate1_single_kernel)
        float_type = np.float32
    else:
        kernel_functions = (MeasureGate0_double_kernel, MeasureGate1_double_kernel)
        float_type = np.float64

    # Apply to state vector
    _0 = random.random()
    _1 = _0 > prob
    if not _1:
        alpha = float_type(1 / np.sqrt(prob))
        kernel_functions[0](
            (block_num, ),
            (thread_per_block, ),
            (index, alpha, vec)
        )
    else:
        alpha = float_type(1 / np.sqrt(1 - prob))
        kernel_functions[1](
            (block_num,),
            (thread_per_block,),
            (index, alpha, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()

    return _1


def apply_resetgate(index, vec, vec_bit, prob, sync: bool = False):
    """
    Measure Gate Measure.
    """
    # Kernel function preparation
    task_number = 1 << (vec_bit - 1)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block
    if vec.dtype == np.complex64:
        kernel_functions = (ResetGate0_single_kernel, ResetGate1_single_kernel)
    else:
        kernel_functions = (ResetGate0_double_kernel, ResetGate1_double_kernel)

    # Apply to state vector
    alpha = np.float64(np.sqrt(prob))
    if alpha < 1e-6:
        kernel_functions[1](
            (block_num, ),
            (thread_per_block,),
            (index, alpha, vec)
        )
    else:
        kernel_functions[0](
            (block_num,),
            (thread_per_block,),
            (index, alpha, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


kernel_funcs = list(locals().keys())
for name in kernel_funcs:
    if name.endswith("kernel"):
        locals()[name].compile()



import os
import yaml


class JobCreator:
    def __init__(self, name: str, circuit_path: str, device: str = "CPU", output_path: str = None):
        self._job_dict = {
            "job_name": name,
            "device": device,
            "circuit": circuit_path,
            "output_path": os.path.abspath('.') if output_path is None else output_path
        }

    def set_simulation_spec(self, shots: int = 100, precision: str = "double", backend: str = "state_vector"):
        simulation_dict = {
            "shots": shots,
            "precision": precision,
            "backend": backend
        }
        self._job_dict["simulation"] = simulation_dict

    def set_qcda_spec(
        self,
        methods: list,
        instruction_set: str = "Google",
        auto_mode: str = "light",
        para: bool = True,
        depara: bool = False,
        templates: dict = None,
        layout_path: str = None
    ):
        qcda_dict = {
            "methods": methods,
            "instruction_set": instruction_set,
            "auto_mode": auto_mode,
            "para": para,
            "depara": depara
        }
        if templates is not None:
            qcda_dict["templates"] = templates

        if layout_path is not None:
            qcda_dict["mapping"] = {"enable": True, "layout_path": layout_path}

        self._job_dict["qcda"] = qcda_dict

    @property
    def job_dict(self):
        return self._job_dict

    def to_yaml(self, output_path: str = None):
        if output_path is None:
            job_name = self._job_dict['job_name']
            output_path = os.path.join(os.path.abspath('.'), f"{job_name}.yml")

        with open(output_path, 'w') as save_file:
            save_file.write(yaml.dump(self._job_dict, allow_unicode=True))



import os
import psutil
import subprocess
import shutil
from typing import Union, Dict

from QuICT.tools import Logger
from QuICT.tools.logger import LogFormat
from QuICT.tools.cli.utils import JobValidation
from .utils import SQLManager


logger = Logger("Job_Management_Local_Mode", LogFormat.full)


class QuICTLocalManager:
    """ QuICT Job Management for the Local Mode. Using SQL to store running-time information. """
    def __init__(self):
        self._sql_connect = SQLManager()

    def _name_validation(self, name) -> bool:
        if not self._sql_connect.job_validation(name):
            logger.warn(
                f"Unmatched job name {name} in local jobs, please using \'quict local job list\' first."
            )
            return False

        # Update given job's status
        self._update_job_status(name)
        return True

    def _job_prepare(self, output_path: str, circuit_path: str):
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        shutil.copyfile(circuit_path, os.path.join(output_path, "circuit.qasm"))

    def start_job(self, job_file: Union[str, Dict]):
        """ Start the job describe by the given yaml file.

        Args:
            job_file (str|dict): The given job's file path or job information dict
        """
        # Validation job_files
        job_info = JobValidation().job_validation(job_file)

        # Check job name
        name = job_info["job_name"]
        if self._sql_connect.job_validation(name):
            logger.warn("Repeated name in local jobs.")
            return

        try:
            self._start_job(name, job_info)
        except Exception as e:
            logger.warn(f"Failure to start job, due to {e}.")

    def _start_job(self, job_name: str, yml_dict: dict):
        # Get information from given yml dict
        job_options = {}
        job_options["circuit_path"] = yml_dict["circuit"]
        job_options["output_path"] = yml_dict["output_path"]
        self._job_prepare(job_options["output_path"], job_options["circuit_path"])

        script_name = ""
        if "simulation" in yml_dict.keys():
            script_name = "simulation"
            job_options["device"] = yml_dict["device"]
            job_options.update(yml_dict['simulation'])

        if "qcda" in yml_dict.keys():
            script_name = "qcda" if script_name == "" else "mixed_pipe"
            job_options.update(yml_dict['qcda'])

        # Pre-paration job's runtime arguments
        runtime_args = ""
        for key, value in job_options.items():
            runtime_args += f"{key}={value} "

        # Start job
        command_file_path = os.path.join(
            os.path.dirname(__file__),
            "../script",
            f"{script_name}.py"
        )
        proc = subprocess.Popen(
            f"python {command_file_path} {runtime_args}", shell=True
        )

        # Save job information into SQL DB.
        job_info = {
            'name': job_name,
            'status': 'running',
            'pid': proc.pid
        }
        self._sql_connect.add_job(job_info)

        logger.info(f"Successfully start the job {job_name} in local mode.")

    def _update_job_status(self, name: str):
        """ Update job's running status, do not distinguish error and finish.

        Args:
            name (str): job's name
        """
        job_status = self._sql_connect.get_job_status(name)
        if job_status in ["finish", "stop"]:
            return

        job_pid = self._sql_connect.get_job_pid(name)
        try:
            _ = psutil.Process(job_pid)
        except:
            self._sql_connect.change_job_status(name, 'finish')

    def status_job(self, name: str):
        """ Get job's states """
        # Check job's list contain given job
        if self._name_validation(name):
            logger.info(
                f"The job state of {name} is {self._sql_connect.get_job_status(name)}"
            )

    def stop_job(self, name: str):
        """ Stop a job. """
        # Check job's list contain given job
        if not self._name_validation(name):
            return

        # check job status
        job_status = self._sql_connect.get_job_status(name)
        if job_status in ["stop", "finish"]:
            logger.info(f"The job {name} is already {job_status}, no need stop anymore.")
            return

        # stop job's process
        job_pid = self._sql_connect.get_job_pid(name)
        try:
            job_process = psutil.Process(job_pid)
            job_process.suspend()
            self._sql_connect.change_job_status(name, 'stop')
            logger.info(f"Successfully stop the job {name}.")
        except Exception as e:
            logger.warn(f"Failure to stop the job {name}, due to {e}.")

    def restart_job(self, name: str):
        """ Restart a job. """
        # Check job's list contain given job
        if not self._name_validation(name):
            return

        # check job status
        job_status = self._sql_connect.get_job_status(name)
        if job_status != "stop":
            logger.info("Restart only for the stop jobs.")
            return

        # restart job's process
        job_pid = self._sql_connect.get_job_pid(name)
        try:
            job_process = psutil.Process(job_pid)
            job_process.resume()
            self._sql_connect.change_job_status(name, "running")
            logger.info(f"Successfully restart the job {name}.")
        except Exception as e:
            logger.warn(f"Failure to restart the job {name}, due to {e}.")

    def delete_job(self, name: str):
        """ Delete a job. """
        # Check job's list contain given job
        if not self._name_validation(name):
            return

        # check job status
        job_status = self._sql_connect.get_job_status(name)
        # kill job's process, if job is still running
        if job_status in ["stop", "running"]:
            job_pid = self._sql_connect.get_job_pid(name)
            try:
                job_process = psutil.Process(job_pid)
                job_process.kill()
                logger.info(f"Successfully kill the job {name}'s process.")
            except Exception as e:
                logger.warn(f"Failure to kill the job {name}, due to {e}.")

        # Delete from Redis DB
        self._sql_connect.delete_job(name)
        logger.info(f"Successfully delete the job {name}.")

    def list_job(self):
        """ List all jobs. """
        all_jobs = self._sql_connect.list_jobs()
        logger.info(f"There are {len(all_jobs)} jobs in local mode.")
        for job_name, _, pid in all_jobs:
            self._update_job_status(job_name)
            logger.info(
                f"job name: {job_name}, related pid: {pid}, " +
                f"job's state: {self._sql_connect.get_job_status(job_name)}"
            )



from flask import Blueprint

from utils.data_structure import JobOperatorType
from utils.email_sender import send_reset_password_email
from script.requset_validation import request_validation
from script.redis_controller import RedisController
from script.sql_controller import SQLManger
from utils.file_manage import create_user_folder
from utils.get_config import get_default_user_config


env_blueprint = Blueprint(name="env", import_name=__name__)
URL_PREFIX = "/quict/env"


@env_blueprint.route(f"{URL_PREFIX}/login", methods=["POST"])
@request_validation(login=True)
def login(**kwargs):
    """ Login user
    kwargs = {
        'json_dict': dict = {
            'username': str,
            'password': str
            }
        }
    """
    json_dict = kwargs['json_dict']
    username = json_dict['username']
    password = json_dict['password']

    return SQLManger().validation_password(username, password)


@env_blueprint.route(f"{URL_PREFIX}/register", methods=["POST"])
@request_validation(register=True)
def register(**kwargs):
    """ Register new user.
    kwargs = {
        'json_dict': dict = {
            'username': str,
            'password': str,
            'email': str,
            }
        }
    """
    json_dict = kwargs['json_dict']
    username = json_dict['username']

    # Create user folder
    create_user_folder(username)

    # Update user info for SQL and Redis
    SQLManger().add_user(json_dict)
    RedisController().update_user_dynamic_info(username, get_default_user_config(username))

    return True


@env_blueprint.route(f"{URL_PREFIX}/unsubscribe", methods=["POST"])
@request_validation()
def unsubscribe(username, **kwargs):
    """ Delete an user. """
    redis_controller = RedisController()
    job_list = redis_controller.list_jobs(username, name_only=True)
    for job_name in job_list:
        redis_controller.add_operator(job_name, JobOperatorType.delete)

    # Delete user in Redis, need to wait all jobs delete first.
    redis_controller.add_operator(username, JobOperatorType.user_delete)

    # Delete user information in database
    SQLManger().delete_user(username)

    return True


@env_blueprint.route(f"{URL_PREFIX}/update_password", methods=["POST"])
@request_validation()
def update_password(username, new_password):
    """ Update user's password. """
    SQLManger().update_password(username, new_password)


@env_blueprint.route(f"{URL_PREFIX}/update_user_info", methods=["POST"])
@request_validation()
def update_user_info(username, new_email: str = None, new_level: int = None):
    """ Update user's information. """
    if new_email is not None:
        SQLManger().update_user_email(username, new_email)

    if new_level is not None:
        SQLManger().update_user_level(username, new_level)


@env_blueprint.route(f"{URL_PREFIX}/forget_password", methods=["POST"])
@request_validation()
def forget_password(username: str, email: str):
    """ Send email for user for activate new password. """
    user_info = SQLManger().get_user_info(username)
    user_email = user_info[1]
    if user_email != email:
        raise KeyError("Unmatched Email address with user.")

    # Send email to user
    reset_password = send_reset_password_email(user_email)
    SQLManger().update_password(username, reset_password)



import numpy as np

from QuICT.algorithm import Algorithm
from QuICT.core import Circuit
from QuICT.core.gate import X, H, Measure
from QuICT.core.gate.backend import MCTOneAux
from QuICT.qcda.synthesis.quantum_state_preparation import QuantumStatePreparation
from QuICT.simulation.state_vector import ConstantStateVectorSimulator


def weight_decison_para(n, k, l):
    kap = k / n
    lam = l / n
    for d in range(n + 2):
        for gamma in range(2 * d - 2):
            s = (1 - np.cos(gamma * np.pi / (2 * d - 1))) / 2
            t = (1 - np.cos((gamma + 1) * np.pi / (2 * d - 1))) / 2
            if lam * s >= kap * t and (1 - kap) * (1 - t) >= (1 - lam) * (1 - s):
                a = np.sqrt((l - k) / (t - s) - (l * s - k * t) / (t - s) - n)
                b = np.sqrt(abs(l * s - k * t) / (t - s))
                return d, gamma, a, b


class WeightDecision(Algorithm):
    @classmethod
    def run(cls, n, k, l, oracle):
        """ decide function f by k-l algorithm by custom oracle
        https://arxiv.org/abs/1801.05717

        Args:
            f(list<int>): the function to be decided
            n(int): the length of function
            k(int): the smaller weight
            l(int): the bigger weight
            oracle(Gate/CompositeGate): the oracle
        Returns:
            int: the ans, k or l
        """
        num = int(np.ceil(np.log2(n + 2))) + 2

        # Determine number of qreg
        circuit = Circuit(num)
        d, gamma, a, b = weight_decison_para(n, k, l)

        # start the eng and allocate qubits
        qreg = circuit[[i for i in range(num - 2)]]
        ancilla = circuit[num - 2]

        # Start with qreg in equal superposition and ancilla in |->
        N = np.power(2, num - 2)
        value = [0 for _ in range(N)]
        for i in range(n):
            value[i] = 1 / np.sqrt(n + a ** 2 + b ** 2)

        value[N - 2] = a / np.sqrt(n + a ** 2 + b ** 2)
        value[N - 1] = b / np.sqrt(n + a ** 2 + b ** 2)

        # Apply oracle U_f which flips the phase of every state |x> with f(x) = 1
        QSP = QuantumStatePreparation('uniformly_gates')
        gates_preparation = QSP.execute(value)
        gates_preparation | circuit(qreg)
        X | circuit(ancilla)
        H | circuit(ancilla)

        MCTOA = MCTOneAux()
        gates_mct = MCTOA.execute(num)

        for i in range(d - 1):
            oracle | circuit([i for i in range(num - 1)])
            gates_mct | circuit
            gates_preparation ^ circuit(qreg)
            for q in qreg:
                X | circuit(q)

            gates_mct | circuit
            for q in qreg:
                X | circuit(q)

            gates_preparation | circuit(qreg)

        # Apply H,X to recover ancilla
        H | circuit(ancilla)
        X | circuit(ancilla)
        oracle | circuit(list(range(num - 1)))
        gates_mct | circuit

        # Measure
        for i in range(num - 1):
            Measure | circuit(i)

        simulator = ConstantStateVectorSimulator()
        _ = simulator.run(circuit)

        if int(ancilla) == gamma % 2:
            return k
        else:
            return l



from typing import *
import numpy as np

from QuICT.core import Circuit
import QuICT.ops.linalg.cpu_calculator as CPUCalculator
from QuICT.tools.exception.core import ValueError
from QuICT.tools.exception.simulation import (
    UnitaryMatrixUnmatchedError, StateVectorUnmatchedError, SampleBeforeRunError
)


class UnitarySimulator():
    """ Algorithms to calculate the unitary matrix of a quantum circuit, and simulate.

    Args:
        device (str, optional): The device type, one of [CPU, GPU]. Defaults to "CPU".
        precision (str, optional): The precision for the unitary matrix, one of [single, double]. Defaults to "double".
    """
    @property
    def vector(self):
        return self._vector

    @vector.setter
    def vector(self, vec):
        self._vector = self._array_helper.array(vec)

    def __init__(
        self,
        device: str = "CPU",
        precision: str = "double"
    ):
        self._device = device
        self._precision = np.complex128 if precision == "double" else np.complex64
        self._vector = None
        self._circuit = None

        if device == "CPU":
            self._computer = CPUCalculator
            self._array_helper = np
        elif device == "GPU":
            import cupy as cp
            import QuICT.ops.linalg.gpu_calculator as GPUCalculator

            self._computer = GPUCalculator
            self._array_helper = cp
        else:
            raise ValueError("UnitarySimulation.device", "[CPU, GPU]", device)

    def initial_vector_state(self):
        """ Initial the state vector for simulation through UnitarySimulator,
        must after initial_circuit
        """
        self._vector = self._array_helper.zeros(1 << self._qubits_num, dtype=self._precision)
        if self._device == "CPU":
            self._vector[0] = self._precision(1)
        else:
            self._vector.put(0, self._precision(1))

    def run(
        self,
        circuit: Union[np.ndarray, Circuit],
        state_vector: np.ndarray = None,
        use_previous: bool = False
    ) -> np.ndarray:
        """ Simulation by given unitary matrix or circuit

        Args:
            circuit (Union[np.ndarray, Circuit]): The unitary matrix or the circuit for simulation
            state_vector (ndarray): The initial state vector.
            use_previous (bool, optional): whether using previous state vector. Defaults to False.

        Returns:
            np.ndarray: The state vector after simulation
        """
        # Step 1: Generate the unitary matrix of the given circuit
        if isinstance(circuit, Circuit):
            if self._precision != circuit._precision:
                circuit.convert_precision()

            self._qubits_num = circuit.width()
            self._unitary_matrix = circuit.matrix(self._device)
            assert 2 ** self._qubits_num == self._unitary_matrix.shape[0], \
                UnitaryMatrixUnmatchedError("The unitary matrix should has the same qubits with the circuit.")
        else:
            row, col = circuit.shape
            self._qubits_num = int(np.log2(row))
            assert row == col and 2 ** self._qubits_num == col, \
                UnitaryMatrixUnmatchedError("The unitary matrix should be square.")
            self._unitary_matrix = self._array_helper.array(circuit, dtype=self._precision)

        # Step 2: Prepare the state vector
        if state_vector is not None:
            assert 2 ** self._qubits_num == state_vector.size, \
                StateVectorUnmatchedError("The state vector should has the same qubits with the circuit.")
            self.vector = self._array_helper.array(state_vector, dtype=self._precision)
        elif not use_previous or self._vector is None:
            self.initial_vector_state()

        # Step 3: Simulation with the unitary matrix and qubit's state vector
        if not self._is_identity():
            self._vector = self._computer.dot(
                self._unitary_matrix,
                self._vector
            )

        return self._vector

    def _is_identity(self):
        identity_matrix = self._array_helper.identity(1 << self._qubits_num, dtype=self._precision)
        return self._array_helper.allclose(self._unitary_matrix, identity_matrix)

    def sample(self, shots: int):
        """_summary_

        Args:
            shots (int): _description_

        Returns:
            _type_: _description_
        """
        assert (self._vector is not None), \
            SampleBeforeRunError("UnitarySimulation sample without run any circuit.")

        original_sv = self._vector.copy()
        counts = [0] * (1 << self._qubits_num)
        for _ in range(shots):
            measured_result = 0
            for i in range(self._qubits_num):
                measured_result <<= 1
                measured_result += self._measure(i)

            counts[measured_result] += 1
            self._vector = original_sv.copy()

        return counts

    def _measure(self, index):
        if self._device == "CPU":
            result = self._computer.measure_gate_apply(
                index,
                self._vector
            )
        else:
            from QuICT.ops.gate_kernel import apply_measuregate, measured_prob_calculate
            prob = measured_prob_calculate(
                index,
                self._vector,
                self._qubits_num
            )
            result = apply_measuregate(
                index,
                self._vector,
                self._qubits_num,
                prob=prob.get()
            )

        return int(result)



import numpy as np

from functools import reduce
from math import gcd
from QuICT.core.gate import X
from QuICT.simulation.state_vector import CircuitSimulator

from QuICT.tools import Logger
from QuICT.tools.exception.core import *

logger = Logger("Shor-util")


def ex_gcd(a, b, arr):
    if b == 0:
        arr[0] = 1
        arr[1] = 0
        return a
    g = ex_gcd(b, a % b, arr)
    t = arr[0]
    arr[0] = arr[1]
    arr[1] = t - int(a / b) * arr[1]
    return g


def mod_reverse(a, n):
    arr = [0, 1]
    g = ex_gcd(a, n, arr)
    if g != 1:
        raise ValueError(f"imput {a} and {n} does not coprime")
    return (arr[0] % n + n) % n


def fast_power(a, b, N):
    x = 1
    now_a = a
    while b > 0:
        if b % 2 == 1:
            x = x * now_a % N
        now_a = now_a * now_a % N
        b >>= 1
    return x


# transform an integer to n-length bitwise string


def int2bitwise(c, n):
    """
    Transform an integer c to binary n-length bitwise string.
    """
    c_bitwise = bin(c)[2:]
    if len(c_bitwise) > n:
        c_bitwise = c_bitwise[-n:]
        # print('c exceeds the length of a, thus is truncated')
    else:
        c_bitwise = "0" * (n - len(c_bitwise)) + c_bitwise
    return c_bitwise


def set(qreg, N):
    """
    Set the qreg as N, using X gates on specific qubits.
    """
    str = bin(N)[2:]
    n = len(qreg)
    m = len(str)
    if m > n:
        logger.warning(
            f"When set qureg as N={N}, N exceeds the length of qureg n={n}, thus is truncated"
        )

    for i in range(min(n, m)):
        if str[m - 1 - i] == "1":
            X | qreg[n - 1 - i]


def split_invert(n, d, CFE):
    CFE.append(n // d)
    n = n % d
    if n == 1:
        CFE.append(d)
        return
    split_invert(d, n, CFE)


def continued_fraction_expansion(n, d):
    """
    Calculate the continued fraction expansion of a rational number n/d.

    Args:
        n: numerator.
        d: denominator.
    """
    CFE = []
    split_invert(n, d, CFE)
    return CFE


def miller_rabin(num):
    """random prime test
        return True, num is a prime whp
        return False, num is a composite
    """
    Test = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]
    if num == 1:
        return False
    t = num - 1
    k = 0
    while (t & 1) == 0:
        k += 1
        t >>= 1
    for test_num in Test:
        # test_num should be generated randomly
        if num == test_num:
            return True
        a = fast_power(test_num, t, num)
        nxt = a
        for _ in range(k):
            nxt = (a * a) % num
            if nxt == 1 and a != 1 and a != num - 1:
                return 0
            a = nxt
        if a != 1:
            return False
    return True


MAX_ROUND = 2


def reinforced_order_finding_constructor(order_finding):
    def reinforced_order_finding(a: int, N: int, eps: float = 1 / 10, simulator=None):
        r_list = []
        i = 0
        while i < MAX_ROUND:
            i += 1
            r = order_finding(a, N, eps, simulator)
            if r != 0 and (a ** r) % N == 1:
                logger.info("\tsuccess!")
                r_list.append(r)
        if len(r_list) == 0:
            return 0
        else:
            return reduce(lambda x, y: (x * y) // gcd(x, y), r_list)

    return reinforced_order_finding


def run_twice_order_finding_constructor(order_finding):
    def run(
        a: int,
        N: int,
        demo: str = None,
        eps: float = 1 / 10,
        simulator=CircuitSimulator(),
    ):
        r1 = order_finding(a, N, eps, simulator)
        r2 = order_finding(a, N, eps, simulator)
        flag1 = pow(a, r1, N) == 1 and r1 != 0
        flag2 = pow(a, r2, N) == 1 and r2 != 0
        if flag1 and flag2:
            r = min(r1, r2)
        elif not flag1 and not flag2:
            r = int(np.lcm(r1, r2))
        else:
            r = int(flag1) * r1 + int(flag2) * r2

        if pow(a, r, N) == 1 and r != 0:
            msg = f"\torder_finding found candidate order: r = {r} of a = {a}"
        else:
            r = 0
            msg = "\torder_finding failed"
        logger.info(msg)
        return r

    return run



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/08/13 15:30
# @Author  : Xiaoquan Xu
# @File    : fermion_operator.py

"""
A fermion operator is a polynomial of anti-commutative creation-annihilation operators,
which is a useful representation for states and Hamiltonians by second quantization.
"""

from QuICT.chemistry.operator.polynomial_operator import PolynomialOperator


class FermionOperator(PolynomialOperator):
    """
    A fermion operator is a polynomial of anti-commutative creation-annihilation operators,
    which is a useful representation for states and Hamiltonians by second quantization.

    Due to the anti-commutation relation, the polynomial is in fact a multilinear function of
    the ladder operators. In this class, the operator could be represented as below.
    For example, list
    [[[(i, 1), (j, 1), (k, 0), (l, 0)], 1.2], [[(k, 1), (l, 1), (i, 0), (j, 0)], -1.2], ...]
    stands for (1.2 a_i^dagger a_j^dagger a_k a_l - 1.2 a_k^dagger a_l^dagger a_i a_j + ...),
    which could also be parsed as string '1.2 * i^ j^ k l - 1.2 * k^ l^ i j + ...'.

    In the following descriptions, the above list is called list format,
    while the above string is called string format.
    """
    def __init__(self, monomial=None, coefficient=1.):
        """
        Create a monomial of ladder operators with the two given formats.

        Args:
            monomial(list/str): Operator monomial in list/string format
            coefficient(int/float/complex): Coefficient of the monomial
        """
        super().__init__(monomial, coefficient)
        if self.operators == []:
            return
        variables = self.operators[0][0]
        l = len(variables)

        # The variables in a monomial should be in ascending order.
        # Anti-commutation relation for operators on different targets.
        for i in range(l - 1, 0, -1):
            fl = False
            for j in range(i):
                if variables[j][0] > variables[j + 1][0]:
                    variables[j], variables[j + 1] = variables[j + 1], variables[j]
                    coefficient *= -1
                    fl = True
            if not fl:
                break

        # Adjacent identical operators lead to zero operator.
        if any([variables[i] == variables[i + 1] for i in range(l - 1)]):
            self.operators = []
            return

        # Anti-commutation relation for operators on identical targets.
        operators = []
        for i in range(l):
            if i == 0 or variables[i][0] != variables[i - 1][0]:
                k = l
                for j in range(i + 1, l):
                    if variables[j][0] != variables[i][0]:
                        k = j
                        break
                if (k - i) % 2 == 1:
                    operators += [variables[i]]
                elif variables[i][1] == 1:
                    operators += [variables[i], variables[i + 1]]
                else:
                    self.operators = [[operators, coefficient],
                                      [operators + [variables[i + 1], variables[i]], -coefficient]]
                    self *= FermionOperator(variables[k:])  # mul promises the ascending order
                    return
        self.operators = [[operators, coefficient]]

    @classmethod
    def get_polynomial(cls, monomial=None, coefficient=1.):
        '''
        Construct an instance of the same class(i.e. FermionOperator) using the arguments.

        Args:
            monomial(list/str): Operator monomial in list/string format
            coefficient(int/float/complex): Coefficient of the monomial
        '''
        return FermionOperator(monomial, coefficient)

    def analyze_single(self, single_operator):
        """
        Transform a string format of a single operator to a tuple.
        For example,
        '231' -> (231,0); '426^' -> (426,1)

        Args:
            single_operator(str): string format

        Returns:
            tuple: the corresponding tuple in list format
        """
        if single_operator[-1] == '^':
            return (int(single_operator[:-1]), 1)
        else:
            return (int(single_operator), 0)

    @classmethod
    def parse_single(cls, single_operator):
        """
        Transform a tuple format of a single operator to a string.
        For example,
        (21,0) -> '21 '; (88,1) -> '88^ '

        Args:
            single_operator(tuple): list format

        Returns:
            string: the corresponding string format
        """
        return str(single_operator[0]) + ('^ ' if single_operator[1] else ' ')



import itertools
from typing import List

from .backward_match import BackwardMatch
from .forward_match import ForwardMatch
from .matching_dag_circuit import Match, MatchingDAGCircuit


class TemplateMatching:
    """
    [1] Iten, R., Moyard, R., Metger, T., Sutter, D. and Woerner, S., 2020.
    Exact and practical pattern matching for quantum circuit optimization.
    `arXiv:1909.05270 <https://arxiv.org/abs/1909.05270>`
    """

    @classmethod
    def _qubit_fixing(cls,
                      circuit: MatchingDAGCircuit,
                      template: MatchingDAGCircuit,
                      c_node_id: int,
                      t_node_id: int,
                      cnt: int):
        """
        Heuristic qubit algorithm. It will explore `cnt` more gate (if possible)
        and return the set of qubits these gates act on.
        """

        t_successors = template.all_successors(t_node_id)
        c_successors = circuit.all_successors(c_node_id)

        if len(t_successors) > (template.size - t_node_id - 1) / 2:
            # explore backwards
            cands = sorted(c_successors)
        else:
            # explore forwards
            cands = sorted(set(range(circuit.size)) - set(c_successors), reverse=True)

        ret = set(circuit.get_node(c_node_id).qargs)
        for cand in cands[: cnt]:
            cur = ret | set(circuit.get_node(cand).qargs)
            if len(cur) > template.width:
                break
            ret = cur
        return list(ret)

    @classmethod
    def execute(cls,
                circuit: MatchingDAGCircuit,
                template: MatchingDAGCircuit,
                qubit_fixing_param: List[int] = None,
                prune_param: List[int] = None
                ):
        """
        Execute the template matching algorithm. It match `circuit` with `template` and return
        all maximal matches start from every qubit as a list.

        Heuristic qubit parameters `qubit_fixing_param` is in the form [cnt] where `cnt` is
        the number of additional qubits explored when enumerating the qubit mapping (recommended
        value is 1).

        Heuristic backward match parameter `prune_param` is in the form [D, W]. Backward match
        will prune the search tree when depth=k*D (k = 1, 2, ...) and at most W maximal matching
        scenarios will survive (recommended value is [3, 1]).

        Above two heuristic algorithms will be executed only when the corresponding parameter is
        specified.

        Args:
            circuit(MatchingDAGCircuit): the circuit to match
            template(MatchingDAGCircuit): the template to be matched
            qubit_fixing_param(List[int]): Heuristic qubit parameters
            prune_param(List[int]): Heuristic backward match parameter

        Returns:
            List[Match]: List of matches
        """

        match_set = set()

        # enumerate the starting node of template
        for t_node_id in range(template.size):
            # enumerate the starting node of circuit
            for c_node_id in range(circuit.size):
                t_node = template.get_node(t_node_id)
                c_node = circuit.get_node(c_node_id)
                if not t_node.compare_with(c_node):
                    continue

                # heuristically fixed qubits
                fixed_q = cls._qubit_fixing(circuit, template, c_node_id, t_node_id, *qubit_fixing_param) \
                    if qubit_fixing_param else []

                # enumerate free qubits
                all_free_q = set(range(circuit.width)) - set(fixed_q)
                for free_q in itertools.combinations(all_free_q, template.width - len(fixed_q)):
                    q_set = list(free_q) + fixed_q
                    # enumerate permutation
                    for mapping in itertools.permutations(q_set, template.width):
                        if not t_node.compare_with(c_node, mapping):
                            continue

                        forward_match = ForwardMatch.execute(
                            circuit, template, c_node_id, t_node_id, list(mapping))

                        match_list = BackwardMatch.execute(
                            circuit, template, forward_match, c_node_id, t_node_id, list(mapping), prune_param)
                        match_set.update(match_list)

        return list(match_set)



from collections import namedtuple
from collections.abc import Iterable
from functools import cached_property
from typing import List, Set

from QuICT.core import Circuit
from QuICT.core.circuit.dag_circuit import DAGCircuit, DAGNode
from QuICT.core.gate import BasicGate

NodeInfo = namedtuple('NodeInfo', ['matched_with', 'is_blocked'])


class MatchingDAGNode(DAGNode):
    """
    DAG node class tailored for template matching algorithm.
    """

    __slots__ = ['successors_to_visit', 'matched_with', 'is_blocked',
                 '_id', '_gate', '_name', '_cargs', '_targs', '_qargs',
                 '_successors', '_predecessors']

    def __init__(self, id: int, gate: BasicGate):
        self.successors_to_visit = []
        self.matched_with = None
        self.is_blocked = False
        super().__init__(id, gate)

    def pop_successors_to_visit(self):
        """
        Pop the first element of successors_to_visit. If no more, return None.

        Returns:
            int: first element of successors_to_visit
        """

        if not self.successors_to_visit:
            return None

        ret = self.successors_to_visit[0]
        self.successors_to_visit.pop(0)
        return ret

    def matchable(self):
        """
        Returns:
            bool: Whether the node can be matched
        """
        return (not self.is_blocked) and (self.matched_with is None)

    def compare_with(self, other, qubit_mapping=None) -> bool:
        """
        Compare `self` to the node `other` under mapping `qubit_mapping`.
        qubit i of `self` is mapped to qubit qubit_mapping[i] of `other`.
        If `qubit_mapping` is None, only compare gate type.
        The order of control qubits is ignored.

        Args:
            other(MatchingDAGNode): the other node
            qubit_mapping(list): mapping
        """

        if self.name != other.name:
            return False
        if qubit_mapping is not None:
            t_cargs = set(map(lambda x: qubit_mapping[x], self.cargs))
            c_cargs = set(other.cargs)
            if t_cargs != c_cargs:
                return False

            for t_qubit, c_qubit in zip(self.targs, other.targs):
                if qubit_mapping[t_qubit] != c_qubit:
                    return False
        return True

    def node_info(self):
        """
        Return:
            tuple: (the matched node, whether is blocked)
        """
        return NodeInfo(self.matched_with, self.is_blocked)


class MatchingDAGCircuit(DAGCircuit):
    """
        DAG circuit class tailored for template matching algorithm.
    """

    def __init__(self, circuit: Circuit):
        self._successor_cache = {}
        self._predecessor_cache = {}
        super().__init__(circuit, node_type=MatchingDAGNode)

    def init_forward_matching(self, node_id, other_id, s2v_enabled=False):
        """
        Initialize forward matching info.
        Args:
            node_id(int): the start node of forward matching
            other_id(int): the start node of the other circuit to match
            s2v_enabled(bool): whether initialize successors_to_visit
        """
        for nid in self.nodes():
            node = self.get_node(nid)
            if node.id == node_id:
                node.matched_with = other_id
                if s2v_enabled:
                    node.successors_to_visit = sorted(node.successors.copy())
            else:
                node.matched_with = None
                if s2v_enabled:
                    node.successors_to_visit = []
            node.is_blocked = False

    def all_predecessors(self, start, cache_enabled=True) -> Set[int]:
        """
        Return all predecessors (direct and undirect) of `start`.
        `start` can be a node id (int) or many node ids (Iterable).
        Return values of single node query will be cached if `cache_enable` is True

        Args:
            start(int/Iterable): the start node(s)
            cache_enabled(bool): whether use cache

        Returns:
            set: set of predecessors
        """

        if not cache_enabled or isinstance(start, Iterable):
            return self._all_reachable(start, 'predecessors')
        elif isinstance(start, int):
            if start not in self._predecessor_cache:
                self._predecessor_cache[start] = set(self.get_node(start).predecessors)
                for succ in self.get_node(start).predecessors:
                    self._predecessor_cache[start] |= self.all_predecessors(succ)
            return self._predecessor_cache[start]
        else:
            assert False, 'start must be int or iterable objects'

    def all_successors(self, start, cache_enabled=True) -> Set[int]:
        """
        Return all successors (direct and indirect) of `start`.
        `start` can be a node id (int) or many node ids (Iterable).
        Return values of single node query will be cached if `cache_enable` is True.

        Args:
            start(int/Iterable): the start node(s)
            cache_enabled(bool): whether use cache

        Returns:
            Set[int]: set of successors
        """

        if not cache_enabled or isinstance(start, Iterable):
            return self._all_reachable(start, 'successors')
        elif isinstance(start, int):
            if start not in self._successor_cache:
                self._successor_cache[start] = set(self.get_node(start).successors)
                for succ in self.get_node(start).successors:
                    self._successor_cache[start] |= self.all_successors(succ)
            return self._successor_cache[start]
        else:
            assert False, 'start must be int or iterable objects'

    def matching_info(self):
        """
        Returns:
              List[NodeInfo]: matching info used by backward_match
        """
        return [self.get_node(i).node_info() for i in range(self.size)]

    def get_circuit(self):
        """
        Output the circuit of this DAG.

        Returns:
            Circuit: the circuit
        """

        circ = Circuit(self.width)
        for node_id in self.nodes():
            node: MatchingDAGNode = self.get_node(node_id)
            circ.append(node.gate.copy())
        return circ


class Match:
    """
    Class to store matches
    """
    def __init__(self, match: List, qubit_mapping: List):
        self.match = sorted(match)
        self.qubit_mapping = qubit_mapping

    def __len__(self):
        return len(self.match)

    def __hash__(self):
        return hash(tuple(self.match))

    def __str__(self):
        return f'Match({str(self.match)}, {str(self.qubit_mapping)})'

    def __repr__(self):
        return str(self)

    @cached_property
    def template_nodes(self):
        """
        Returns:
            Set[int]: Set of tempalte nodes
        """

        return {m[0] for m in self.match}

    @cached_property
    def circuit_nodes(self):
        """
        Returns:
            Set[int]: Set of circuit nodes
        """

        return {m[1] for m in self.match}



import os
import yaml
from typing import Union, Dict

from QuICT.core.layout import Layout
from QuICT.tools.interface import OPENQASMInterface


class JobValidation:
    _SIMULATION_PARAMETERS = {
        "precision": ["single", "double"],
        "backend": ["state_vector", "density_matrix", "unitary"],
        "device": ["CPU", "GPU"]
    }

    _BASED_METHODS = ["GateTransform", "Clifford", "CliffordRz", "Commutative", "SymbolicClifford", "Template", "CNOT"]

    _QCDA_PARAMETERS = {
        "instruction_set": ["USTC", "Google", "IBMQ", "IonQ", "Nam", "Origin"],
        "auto_mode": ["light", "heavy"]
    }

    def __init__(self):
        self._job_template_path = os.path.join(
            os.path.dirname(__file__),
            "../template"
        )

    @staticmethod
    def get_circuit_info(qasm_file: str) -> dict:
        """ whether the given qasm file is valid or not. If not valid, raise ValueError. """
        try:
            circuit = OPENQASMInterface.load_file(qasm_file).circuit
        except:
            raise ValueError(f"Failure to load circuit from given file. {qasm_file}.")

        return {
            "width": str(circuit.width()),
            "size": str(circuit.size()),
            "depth": str(circuit.depth())
        }

    @staticmethod
    def load_yaml_file(file_path: str) -> dict:
        with open(os.path.abspath(file_path), encoding='utf-8') as yml:
            job_info = yaml.load(yml)

        return job_info

    @classmethod
    def based_info_validate(self, job_dict):
        # Necessary feature
        name = job_dict["job_name"]
        assert isinstance(name, str), f"Job's name shoule be a string, not {type(name)}."
        assert job_dict["device"] in self._SIMULATION_PARAMETERS['device'], \
            "Job's resource device should be one of [CPU, GPU]."

        # circuit's qasm file validation
        circuit_info = JobValidation.get_circuit_info(job_dict["circuit"])
        job_dict['circuit_info'] = circuit_info

    @classmethod
    def simulation_validation(self, simulation_dict: dict):
        # validation job's simulation
        assert simulation_dict["shots"] >= 1, "The number of shots >= 1."
        assert simulation_dict["precision"] in self._SIMULATION_PARAMETERS['precision'], \
            "precesion should be one of [single, double]."
        assert simulation_dict["backend"] in self._SIMULATION_PARAMETERS['backend'], \
            "backend should be one of [state_vector, density_matrix, unitary]."

    @classmethod
    def qcda_validation(self, qcda_dict: dict):
        # QCDA methods validation
        methods = qcda_dict["methods"]
        for method in methods:
            assert method in self._BASED_METHODS, \
                f"Unrecognized QCDA method, please use one of {self._BASED_METHODS}."

        # Extra Args Validation
        if "GateTransform" in methods:
            assert qcda_dict["instruction_set"] in self._QCDA_PARAMETERS["instruction_set"], \
                "Wrong Instruction Set."

        if "Auto" in methods:
            assert qcda_dict["auto_mode"] in self._QCDA_PARAMETERS["auto_mode"], \
                "auto_mode should be one of [light, heavy]."

        if "Commutative" in methods:
            assert isinstance(qcda_dict["para"], bool) and isinstance(qcda_dict["depara"], bool), \
                "The para and depara should be bool."

        if "Template" in methods:
            template_dict = qcda_dict["templates"]
            temp_args = [template_dict["max_width"], template_dict["max_size"], template_dict["max_depth"]]
            for arg in temp_args:
                assert isinstance(arg, int)

            qcda_dict["templates"] = "+".join([str(targ) for targ in temp_args])
        else:
            qcda_dict["templates"] = ""

        # Validate qcda mapping args
        if qcda_dict["mapping"]["enable"]:
            try:
                _ = Layout(1).load_file(qcda_dict["mapping"]["layout_path"])
            except Exception as e:
                raise KeyError(f"Failure to load layout from file as {e}.")

            # Update qcda dict
            qcda_dict["layout_path"] = qcda_dict["mapping"]["layout_path"]

        # adjust un-used options
        del qcda_dict["mapping"]
        qcda_dict["methods"] = "+".join(methods)

    def job_complement(self, info: dict, job_type: str):
        default_job_path = os.path.join(self._job_template_path, "quict_job.yml")
        default_job_dict = JobValidation.load_yaml_file(default_job_path)
        core_part = default_job_dict[job_type]

        for key, value in info.items():
            if key in list(core_part.keys()):
                core_part[key] = value

        return core_part

    def job_validation(self, job_file_path: Union[str, Dict]) -> Dict:
        """ Validate the given job's yaml file, and generate the regularized job's file.
        regularized_job_dict = {
            job_name(str),
            type(str): one of [qcda, simulation],
            circuit(str): circuit's qasm file path,
            output_path(str): The output path for store result
            circuit_info:
                qasm(str): circuit's qasm,
                width(int): the number of qubits in circuit,
                size(int): the number of qubits in circuit,
                depth(int): the number of qubits in circuit,

            ### only for simulation
            simulation:
                shots(int),
                precision(str), one of [single, double],
                backend(str), one of [state_vector, density_matrix, unitary]
                device(str), one of [CPU, GPU]

            ### only for qcda
            qcda:
                methods(list): []       # The methods for QCDA
                **extra_args:           # The arguments for QCDA methods

        Args:
            job_file_path (str): The file path for job's yaml file.

        Returns:
            regularized_job_dict(dict): The regularized job's dict
        """
        # Step 1: Load yaml file from given file path
        if isinstance(job_file_path, str):
            job_info = JobValidation.load_yaml_file(job_file_path)
        else:
            job_info = job_file_path

        # Step 2: check based information of job_yaml
        self.based_info_validate(job_info)

        # Step 3: complement and validate job's simulation/qcda information
        if "simulation" in job_info.keys():
            required_info = self.job_complement(job_info["simulation"], "simulation")
            self.simulation_validation(required_info)
            job_info["simulation"] = required_info

        if "qcda" in job_info.keys():
            required_info = self.job_complement(job_info["qcda"], "qcda")
            self.qcda_validation(required_info)
            job_info["qcda"] = required_info

        # Step 4: Prepare output path
        job_info["output_path"] = os.path.join(job_info["output_path"], job_info["job_name"])

        return job_info



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/12/2 下午9:58
# @Author  : Han Yu
# @File    : sabre

import copy
from typing import List, Optional

from QuICT.core.layout import Layout
from QuICT.core.gate import *
from QuICT.core.circuit import Circuit


class DAGNode:
    """
        DAG representation of a quantum circuit that indicates the commutative
        relations between gates.
    """
    def __init__(self, gate):
        """
          Args:
              gate_(BasicGate): Gate represented by this node
        """
        self.gate = gate
        self.bits = gate.cargs + gate.targs
        self.edges = []
        # one qubit gate attach forward the node, it can be executed after this node immediately
        self.attach = []
        self.pre_number = 0


class SABREMapping:
    """
        Mapping with the heuristic algorithm SABRE

        Reference:

        Li G, Ding Y, Xie Y. Tackling the qubit mapping problem for NISQ-era quantum devices[C]
        Proceedings of the Twenty-Fourth International Conference on Architectural Support for
        Programming Languages and Operating Systems. 2019: 1001-1014.
        <https://arxiv.org/abs/1809.02573>

    """
    def __init__(
        self,
        layout: Layout,
        sizeE=20,
        w=0.5,
        epsilon=0.001
    ):
        """
            Args:
                layout(Layout): the layout of the physical quantum circuit
                sizeE(int): the size of the extended set, default 20
                w(float): the weight of the extended set, default 0.5
                epsilon(float): the decay parameter of the SABRE algorithm, default 0.001
        """
        self._layout = layout
        self._sizeE = sizeE
        self._w = w
        self._epsilon = epsilon
        self.phy2logic = None
        self.logic2phy = None

    def execute(
        self,
        circuit: Union[Circuit, CompositeGate],
        initial_l2p: List[int] = None
    ) -> Union[Circuit, CompositeGate]:
        """
            Args:
                circuit (Circuit/CompositeGate): The circuit/CompositeGate to be mapped
                initial_l2p (List[int]): The initial mapping of the circuit, default identity

            Returns:
                Circuit: the executable circuit on the physical device.
        """

        SIZE_E = self._sizeE
        W = self._w
        layout = self._layout
        decay_parameter = self._epsilon
        decay_cycle = 5

        nodes = []
        qubit_number = circuit.width()
        phy_number = layout.qubit_number

        # floyd algorithm, D indicate the distance of qubit on the physical device.
        D: List[List[int]] = []
        for _ in range(phy_number):
            D.append([phy_number for _ in range(phy_number)])

        for edge in layout.edge_list:
            D[edge.u][edge.v] = D[edge.v][edge.u] = 1
        for i in range(phy_number):
            for j in range(phy_number):
                if j == i:
                    continue
                for k in range(phy_number):
                    if k == i or k == j:
                        continue
                    D[i][j] = min(D[i][j], D[i][k] + D[k][j])

        # the first layer of the DAG(live update)
        F = []

        # extend the logic2phy(size equal with physical device) and calculate phy2logic
        if initial_l2p is None:
            l2p = [i for i in range(phy_number)]
        else:
            l2p = copy.deepcopy(initial_l2p)
            goal = set([i for i in range(phy_number)])
            for number in initial_l2p:
                goal.remove(number)
            while len(l2p) < phy_number:
                l2p.append(goal.pop())
        p2l = [0 for _ in range(phy_number)]
        for index in l2p:
            p2l[l2p[index]] = index

        def can_execute(node):
            """
                whether a node in DAG can be executed now.
            """
            if len(node.bits) == 1:
                return True
            elif len(node.bits) == 2:
                return layout.check_edge(l2p[node.bits[0]], l2p[node.bits[1]])
            else:
                assert False

        def obtain_swaps():
            """
                obtain all candidate swap with first layer of the DAG
            """
            candidates = []
            bits = set()
            for node in F:
                if len(node.bits) == 1:
                    continue
                bits = bits.union(set([l2p[bit] for bit in node.bits]))
            for edge in layout.edge_list:
                if edge.u in bits or edge.v in bits:
                    candidates.append(edge)
            return candidates

        def temp_pi(edge):
            """
                Generate a new logic2phy with a swap indicated by a layoutEdge
            """
            new_mapping = copy.deepcopy(l2p)
            u, v = edge.u, edge.v
            new_mapping[p2l[u]] = v
            new_mapping[p2l[v]] = u
            return new_mapping

        def phy_gate(g):
            """
                Mapping a logic gate to a phy gate with logic2phy.
            """
            _g = g.copy()
            _g.cargs = [l2p[carg] for carg in g.cargs]
            _g.targs = [l2p[targ] for targ in g.targs]
            return _g

        def heuristic_cost(newl2p):
            """
                the heuristic_cost function
            """
            H_basic = 0
            H_extend = 0
            F_count = len(F)
            E_queue = []
            for node in F:
                H_basic += D[newl2p[node.bits[0]]][newl2p[node.bits[1]]] / F_count
                E_queue.append(node)

            ESet = []
            decQueue = []
            while len(ESet) < SIZE_E and len(E_queue) > 0:
                node = E_queue.pop(0)
                decQueue.append(node)
                for succ in node.edges:
                    succ.pre_number -= 1
                    assert succ.pre_number >= 0
                    if succ.pre_number == 0:
                        ESet.append(succ)
                        E_queue.append(succ)

            E_count = len(ESet)
            for node in ESet:
                H_extend += D[newl2p[node.bits[0]]][newl2p[node.bits[1]]] / E_count

            for node in decQueue:
                for n in node.edges:
                    n.pre_number += 1

            return H_basic + W * H_extend

        # build the DAG
        exe_gates = []
        predag: List[Optional[DAGNode]] = [None for _ in range(qubit_number)]
        for gate in circuit.gates:
            node = DAGNode(gate)
            nodes.append(node)
            pre_number = 0
            if len(node.bits) == 1:
                dag = predag[node.bits[0]]
                if dag is not None:
                    dag.attach.append(node)
                else:
                    exe_gates.append(phy_gate(node.gate))
            else:
                for bit in node.bits:
                    dag = predag[bit]
                    if dag is not None:
                        if len(node.bits) == 0:
                            dag.attach.append(node)
                        elif node not in dag.edges:
                            dag.edges.append(node)
                            pre_number += 1
                for bit in node.bits:
                    predag[bit] = node

            node.pre_number = pre_number
            if len(node.bits) == 2 and pre_number == 0:
                F.append(node)

        # the main process of the SABRE algorithm
        decay = [1 for _ in range(phy_number)]
        decay_time = 0
        while len(F) > 0:
            decay_time += 1
            if decay_time % decay_cycle == 0:
                decay = [1 for _ in range(phy_number)]
            exe_gate_list = []
            for node in F:
                if can_execute(node):
                    exe_gate_list.append(node)
                    exe_gates.append(phy_gate(node.gate))
                    nodes.remove(node)
                    for gate in node.attach:
                        nodes.remove(gate)
                        exe_gates.append(phy_gate(gate.gate))
            if len(exe_gate_list) != 0:
                for node in exe_gate_list:
                    F.remove(node)
                    for succ in node.edges:
                        succ.pre_number -= 1
                        if succ.pre_number < 0:
                            assert False
                        if succ.pre_number == 0:
                            F.append(succ)
                continue
            else:
                candidate_list = obtain_swaps()
                theSwap = None
                theScore = 0
                thePi = None
                for swap in candidate_list:
                    pi = temp_pi(swap)
                    H_score = heuristic_cost(pi)
                    H_score = H_score * max(decay[p2l[swap.u]], decay[p2l[swap.v]])
                    if theSwap is None or H_score < theScore:
                        theScore = H_score
                        theSwap = swap
                        thePi = pi
                p2l[theSwap.u], p2l[theSwap.v] = p2l[theSwap.v], p2l[theSwap.u]
                l2p = thePi
                exe_gates.append((Swap & ([theSwap.u, theSwap.v])))
                decay[p2l[theSwap.u]] += decay_parameter
                decay[p2l[theSwap.v]] += decay_parameter

        circuit = Circuit(len(l2p))
        for g in exe_gates:
            g | circuit
        self.phy2logic = p2l
        self.logic2phy = l2p
        return circuit

    def execute_initialMapping(
        self,
        circuit: Union[Circuit, CompositeGate],
        initial_l2p: List[int] = None
    ) -> List[int]:
        """
            Args:
                circuit (Circuit/CompositeGate): The circuit/CompositeGate to be mapped
                initial_l2p (List[int]): The initial mapping of the circuit, default identity

            Returns:
                List[int]: the initial mapping by SABRE
        """

        qubit = circuit.width()
        reverse_qc = Circuit(qubit)
        for index in range(len(circuit.gates) - 1, -1, -1):
            _gate = circuit.gates[index].copy()
            _gate | reverse_qc

        self.execute(circuit, initial_l2p=initial_l2p)
        newMP = copy.deepcopy(self.logic2phy)
        self.execute(reverse_qc, initial_l2p=newMP)
        newMP = copy.deepcopy(self.logic2phy)
        return newMP



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/8/22 2:39
# @Author  : Han Yu
# @File    : _algorithm.py


class Algorithm(object):
    """ quantum algorithm which should run in the quantum hardware

    the subClass should overloaded the function run
    """

    @staticmethod
    def run(*pargs):
        """
        Args:
            *pargs: parameters which needed by the algorithm.
        """
        return pargs[0]



import json
import redis

from server.utils.data_structure import JobOperatorType, JobState
from server.utils.get_config import get_default_user_config


class RedisController:
    def __init__(self):
        try:
            self._redis_pool = redis.ConnectionPool(host='127.0.0.1', port=6379, db=0)
            self._redis_connection = redis.Redis(connection_pool=self._redis_pool, decode_responses=True)
        except Exception as e:
            raise KeyError(f"Failure to connect to Redis, due to {e}. please check master node redis connection.")

    ####################################################################
    ############               User DB Function             ############
    ####################################################################
    def get_user_dynamic_info(self, user_name: str):
        return self.dict_decode(
            self._redis_connection.hgetall(f"User_Dynamic_Info:{user_name}")
        )

    def update_user_dynamic_info(self, user_name: str, user_info: dict = None):
        """ Update/Initial User Information in Redis.

        Args:
            user_name (str): The user's name
            user_info (dict, optional): _description_. Defaults to None.
                user_info: dict = {
                    username: str,
                    maximum_parallel_level: int,
                    GPU_allowence: bool,
                    max_qubits_in_cpu: float,
                    max_qubits_in_gpu: float,
                    number_of_running_jobs: int,
                    running_qubits_in_gpu: float,
                    running_qubits_in_cpu: float,
                }
        """
        if user_info is None:
            if self._redis_connection.exists(f"User_Dynamic_Info:{user_name}"):
                return

            user_info = get_default_user_config(user_name)

        self._redis_connection.hmset(f"User_Dynamic_Info:{user_name}", user_info)

    def delete_user_dynamic_info(self, user_name: str):
        self._redis_connection.delete(f"User_Dynamic_Info:{user_name}")

    ####################################################################
    ############              Jobs DB Function              ############
    ####################################################################
    def get_pending_jobs_queue(self):
        return self.list_decode(self._redis_connection.lrange("pending_jobs", 0, -1))

    def get_running_jobs_queue(self):
        return self.list_decode(self._redis_connection.lrange("running_jobs", 0, -1))

    def get_finish_jobs_queue(self):
        return self.list_decode(self._redis_connection.lrange("finish_jobs", 0, -1))

    def get_operator_queue(self):
        return self.list_decode(self._redis_connection.lrange("operator_queue", 0, -1))

    def get_job_info(self, job_name):
        return self.dict_decode(self._redis_connection.hgetall(f"Job_Info:{job_name}"))

    def list_jobs(self, username: str, name_only: bool = False):
        keys = self._redis_connection.keys(f"Job_Info:{username}*")
        if name_only:
            return self.list_decode(keys)

        job_infos_str = []
        for k in keys:
            table = self.dict_decode(self._redis_connection.hgetall(k.decode('utf-8')))
            job_infos_str.append(table)

        return job_infos_str

    def add_pending_job(self, job_dict: dict):
        job_name = job_dict['job_name']
        username = job_dict['username']
        if self._redis_connection.exists(f"Job_Info:{username}:{job_name}"):
            raise KeyError("repeated name.")

        self._redis_connection.rpush("pending_jobs", f"{username}:{job_name}")

        # Add job info into JobInfo Table
        job_dict['state'] = JobState.Pending.value
        self._redis_connection.hmset(f"Job_Info:{username}:{job_name}", job_dict)

    def add_running_job_from_pending_jobs(self, job_name: str):
        self._redis_connection.rpush("running_jobs", job_name)
        self._redis_connection.lrem("pending_jobs", 0, value=job_name)
        self._redis_connection.hset(f"Job_Info:{job_name}", 'state', JobState.Running.value)

    def add_finish_job_from_running_jobs(self, job_name: str, job_state: JobState):
        self._redis_connection.rpush("finish_jobs", job_name)
        self._redis_connection.lrem("running_jobs", 0, value=job_name)
        self._redis_connection.hset(f"Job_Info:{job_name}", 'state', job_state.value)

    def add_operator(self, job_name: str, operator: JobOperatorType):
        if (
            operator.value != JobOperatorType.user_delete.value and
            not self._redis_connection.exists(f"Job_Info:{job_name}")
        ):
            raise KeyError("Try to operate non-existed Job.")

        self._redis_connection.rpush("operator_queue", f"{job_name}:{operator.value}")

    def remove_job(self, job_name: str, job_state: str = None):
        if not self._redis_connection.exists(f"Job_Info:{job_name}"):
            raise KeyError("job not in database.")

        if job_state is None:
            job_state = self._redis_connection.hget(f"Job_Info:{job_name}", "state").decode()

        if job_state == JobState.Error.value:
            job_state = "finish"

        self._redis_connection.lrem(f"{job_state}_jobs", 0, job_name)
        self._redis_connection.delete(f"Job_Info:{job_name}")

    def remove_operator(self, op: str):
        self._redis_connection.lrem('operator_queue', 0, op)

    def change_job_state(self, job_name: str, state: JobState):
        self._redis_connection.hset(f"Job_Info:{job_name}", 'state', state.value)

    ####################################################################
    ############            Redis Utils Function            ############
    ####################################################################
    def flush_all(self):
        self._redis_connection.flushdb()

    def list_decode(self, val: list):
        return [v.decode() for v in val]

    def dict_decode(self, val: dict):
        decode_dict = {}
        for key, value in val.items():
            try:
                value = json.loads(value)
            except:
                value = value.decode()
            finally:
                decode_dict[key.decode()] = value

        return decode_dict



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/5/21 下午3:24
# @Author  : Kaiqi Li
# @File    : calculation_layer

import weakref
import cupy as cp
import numpy as np

import QuICT.ops.linalg.gpu_calculator as GPUCalculator


class CalculationLayer:
    """ The context class of GPU linear algorithm and GPU memory management.

    Args:
        gpu_device (int): Indicated GPU device.
        memory_limit (int): the limitation of memory pool. Unit is GiB

    Examples:
        with CalculationLayer(gpu_device=0, memory_limit=4) as CL:
            gpu_A = cp.array(cpu_A)
            gpu_B = cp.array(cpu_B)

            result = CL.dot(gpu_A, gpu_B, gpu_out=True)
    """

    def __init__(self, gpu_device: int = 0, memory_limit: int = None):
        if gpu_device != cp.cuda.runtime.getDevice() and gpu_device < cp.cuda.runtime.getDeviceCount():
            dev = cp.cuda.Device(gpu_device)
            cp.cuda.Device.use(dev)

        self.mempool = cp.get_default_memory_pool()
        self.pinned_mempool = cp.get_default_pinned_memory_pool()

        if memory_limit:
            self.mempool.set_limit(memory_limit * 1024 ** 3)

        # Set Memory allocator
        cp.cuda.set_allocator(self.mempool.malloc)

        # Memory Record
        self.mempool_used = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clear_memory()

    @property
    def memory_pool(self):
        return self.mempool

    @property
    def pinned_memory_pool(self):
        return self.pinned_mempool

    def clear_memory(self):
        """ released unused memory in current memory pool. """
        self.mempool_used.clear()
        self.mempool.free_all_blocks()
        self.pinned_mempool.free_all_blocks()

    def htod(self, target):
        """ mv target from host into GPU device. """
        if type(target) is not cp.ndarray:
            cp_t = cp.array(target)
            weak_r = weakref.ref(cp_t)
            self.mempool_used.append(cp_t)

            return weak_r

        raise ("The given value has been added in the GPU.")

    def dtoh(self, target):
        """ mv target from GPU device into host. """
        if type(target) is cp.ndarray:
            return target.get()

        if type(target) is weakref.ref:
            return target().get()

        raise ("The given value not in GPU.")

    def _var_normalized(self, target):
        if type(target) is weakref.ref:
            return target()

        if type(target) is np.ndarray:
            wr_target = self.htod(target)
            return wr_target()

        return target

    def _result_normalized(self, result, gpu_out):
        if gpu_out:
            return result
        else:
            self.mempool_used.append(result)
            wr_result = weakref.ref(result)
            return wr_result

    def dot(self, A, B, gpu_out: bool = True, sync: bool = True):
        """ dot matrix A and matrix B

        Args:
            A(np.array<np.complex>): the matrix A
            B(np.array<np.complex>): the matrix B
            gpu_out(bool): return result from GPU into CPU

        Returns:
            np.array<np.complex>: A * B
        """
        gpu_A = self._var_normalized(A)
        gpu_B = self._var_normalized(B)

        result = GPUCalculator.dot(gpu_A, gpu_B, gpu_out, sync)

        return self._result_normalized(result, gpu_out)

    def tensor(self, A, B, gpu_out: bool = True, sync: bool = True):
        """ tensor A and B

        Args:
            A(np.array<np.complex>): the matrix A
            B(np.array<np.complex>): the matrix B
            gpu_out(bool): return result from GPU into CPU

        Returns:
            np.array<np.complex>: the tensor result A ⊗ B
        """
        gpu_A = self._var_normalized(A)
        gpu_B = self._var_normalized(B)

        result = GPUCalculator.tensor(gpu_A, gpu_B, gpu_out, sync)

        return self._result_normalized(result, gpu_out)

    def MatrixTensorI(self, A, n, m, gpu_out: bool = True, sync: bool = True):
        """ tensor I^n and A and I^m

        Args:
            A(np.array<np.complex>): the matrix A
            n(int): the index of indentity
            m(int): the index of indentity
            gpu_out(bool): return result from GPU into CPU

        Returns:
            np.array<np.complex>: the tensor result I^n ⊗ A ⊗ I^m
        """
        gpu_A = self._var_normalized(A)

        result = GPUCalculator.MatrixTensorI(gpu_A, n, m, gpu_out, sync)

        return self._result_normalized(result, gpu_out)

    def VectorPermutation(
            self,
            A,
            mapping,
            changeInput: bool = False,
            gpu_out: bool = True,
            sync: bool = True
    ):
        """ permutaion A with mapping, inplace

        Args:
            A(np.array<np.complex>): the matrix A.
            mapping(np.array<int>): the qubit mapping.
            changeInput(bool): whether changes in A.
            gpu_out(bool): return result from GPU.

        Returns:
            np.array<np.complex>: the result of Permutation
        """
        gpu_A = self._var_normalized(A)

        result = GPUCalculator.VectorPermutation(gpu_A, mapping, changeInput, gpu_out, sync)

        return self._result_normalized(result, gpu_out)

    def MatrixPermutation(
            self,
            A,
            mapping,
            changeInput: bool = False,
            gpu_out: bool = True,
            sync: bool = True
    ):
        """ permute mat with mapping, inplace

        Args:
            A(np.array<np.complex>): the matrix A.
            mapping(np.array<int>): the qubit mapping.
            changeInput(bool): whether changes in A.
            gpu_out(bool): return result from GPU.
        """
        gpu_A = self._var_normalized(A)

        result = GPUCalculator.MatrixPermutation(gpu_A, mapping, changeInput, gpu_out, sync)

        return self._result_normalized(result, gpu_out)



from .backend.mct import *
from QuICT.core.gate import CompositeGate, X, CX, CCX
from QuICT.tools.exception.core import TypeError, ValueError


class MultiControlToffoli(object):
    """
    Divided by the usages of auxiliary qubits, here are 4 implementations of multi-control Toffoli gates.
    """
    __AUX_USAGES = ['no_aux', 'one_clean_aux', 'one_dirty_aux', 'half_dirty_aux']

    def __init__(self, aux_usage='no_aux'):
        """
        Args:
            aux_usage(str): 4 different usages of auxiliary qubits could be chosen, as listed
                'no_aux': No auxiliary qubits are used
                'one_clean_aux': 1 clean auxiliary qubit is used
                'one_dirty_aux': 1 dirty auxiliary qubit is used
                'half_dirty_aux': more than half of all qubits are used as auxiliary qubits, which could be dirty
        """
        assert aux_usage in self.__AUX_USAGES, TypeError("MultiControlToffoli.aux_usage", self.__AUX_USAGES, aux_usage)

        self.aux_usage = aux_usage

    def __call__(self, control, aux=0):
        """
        By default, the qubit arrangement would be control qubits, target qubit, auxiliary qubits (if exists)

        Args:
            control(int): the number of control qubits
            aux(int, optional): only works when 'half_dirty_aux' is chosen, the number of auxiliary qubits

        Returns:
            CompositeGate: the mct gates
        """
        # Special cases, no MCT would be executed.
        if control == 0:
            gates = CompositeGate()
            X & 0 | gates
            return gates
        if control == 1:
            gates = CompositeGate()
            CX & [0, 1] | gates
            return gates
        if control == 2:
            gates = CompositeGate()
            CCX & [0, 1, 2] | gates
            return gates

        # Otherwise
        if self.aux_usage == 'no_aux':
            return MCTWithoutAux().execute(control + 1)
        elif self.aux_usage == 'one_clean_aux':
            return MCTOneAux().execute(control + 2)
        elif self.aux_usage == 'one_dirty_aux':
            return MCTLinearOneDirtyAux().execute(control + 2)
        else:
            qubit = control + aux + 1
            if control > (qubit // 2) + (1 if qubit % 2 == 1 else 0):
                raise ValueError("MultiControlToffoli.control", f"<= ceil({qubit}/2)", control)
            controls = [i for i in range(control)]
            auxs = [i for i in range(control + 1, qubit)]
            return MCTLinearHalfDirtyAux.assign_qubits(qubit, control, controls, auxs, control)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/5/12 下午4:41
# @Author  : Kaiqi Li
# @File    : utils

import numpy as np

from numba import njit


def perm_sort(indexes: np.ndarray, blocks: int):
    n = len(indexes)
    iter = n // blocks
    perm_op = []

    # block level swap
    for i in range(blocks):
        for j in range(i + 1, blocks):
            swap = True
            for z in range(iter):
                if indexes[j * iter + z] // iter != i:
                    swap = False
                    break

            if swap:
                perm_op.append(("ALL", i, j))
                indexes[i * iter:i * iter + iter], indexes[j * iter:j * iter + iter] = \
                    indexes[j * iter:j * iter + iter], indexes[i * iter:i * iter + iter]

    for i in range(n):
        block_interval = i // iter
        if indexes[i] // iter != block_interval:
            for j in range(indexes[i] // iter, indexes[i] // iter + iter):
                if indexes[j] // iter == block_interval:
                    break

            indexes[i], indexes[j] = indexes[j], indexes[i]
            perm_op.append(("IDX", i, j))

    return perm_op, indexes


@njit(nogil=True)
def mapping_augment(mapping: np.ndarray) -> np.ndarray:
    n = len(mapping)
    p2n = 1 << n
    res = np.zeros(shape=p2n, dtype=np.int64)
    for i in range(p2n):
        for k in range(n):
            res[i] |= ((i >> (n - 1 - mapping[k])) & 1) << (n - 1 - k)
    return res



from numpy.random import random
from typing import Union, List

from .noise_error import QuantumNoiseError, NoiseChannel
from .readout_error import ReadoutError
from QuICT.core import Circuit
from QuICT.core.gate import BasicGate, GateType, GATE_TYPE_TO_CLASS, Unitary
from QuICT.core.operator import NoiseGate
from QuICT.tools.exception.core import TypeError, ValueError, NoiseApplyError


class NoiseModel:
    """
    The Noise Model, which contains the QuantumNoiseErrors and can translated circuit with the given QuantumNoiseError.

    Args:
        name (str, optional): The name of this NoiseModel. Defaults to "Noise Model".
        basic_gates (List, optional): The list of quantum gates will apply the NoiseError.
            Defaults to [h, s, sdg, x, y, z, rx, ry, rz, u1, t, tdg, cx, cz, crz, cu1].
    """
    __BASED_GATES = [
        "h", "s", "sdg", "x", "y", "z", "rx", "ry", "rz",
        "u1", "t", "tdg", "cx", "cz", "crz", "cu1"
    ]

    def __init__(self, name: str = "Noise Model", basic_gates: List = __BASED_GATES):
        self._name = name
        self._basic_gates = basic_gates

        self._error_by_gate = {}
        for gate in basic_gates:
            self._error_by_gate[gate] = []

        self._readout_errors = []

    def __str__(self):
        nm_str = f"{self._name}:\nBasic Gates: {self._basic_gates}\nNoise Errors:\n"
        if self._all_qubits_error_gates:
            nm_str += f"Gates with all qubits: {set(self._all_qubits_error_gates)}\n"

        if self._specified_error_gates:
            nm_str += "Gates with specified qubits:"
            for _, q, g in self._instruction["specific"]:
                nm_str += f"[qubits: {q}, gates: {g}] "

        if self._readout_errors:
            nm_str += "Measured Gates with Readout Error:"
            for idx, re_qubit in enumerate(self._readout_errors):
                readout_error = self._instruction["readout"][idx]
                nm_str += f"[qubits: {re_qubit}, Readout Error Prob: {readout_error.prob}]"

        return nm_str

    def _qubits_normalize(self, qubits: Union[int, List[int]]):
        if isinstance(qubits, int):
            qubits = [qubits]
        elif not isinstance(qubits, (list, tuple)):
            raise TypeError("NoiseModel.add.qubits", "int/list<int>", type(qubits))

        for q in qubits:
            if q < 0 or not isinstance(q, int):
                raise ValueError("NoiseModel.add.qubits", "be a positive integer", q)

        return qubits

    def _gates_normalize(self, noise: QuantumNoiseError, gates: Union[str, List[str]]):
        assert isinstance(noise, QuantumNoiseError), \
            TypeError("NoiseModel.add.noise", "QuantumNoiseError", type(noise))
        if not gates:
            raise ValueError("NoiseModel.add.gates", "not be empty", gates)

        if isinstance(gates, str):
            gates = [gates]
        elif not isinstance(gates, (list, tuple)):
            raise TypeError("NoiseModel.add.gates", "str/list<str>", type(gates))

        for g in gates:
            if g not in self._basic_gates:
                raise ValueError("NoiseModel.add.gates", self._basic_gates, g)

            gate = GATE_TYPE_TO_CLASS[GateType.__members__[g]]()
            if gate.controls + gate.targets != noise.qubits:
                raise NoiseApplyError(
                    f"Un-matched qubits number between gate {gate.controls + gate.targets}" +
                    f"with noise error {noise.qubits}."
                )

        return gates

    def add(self, noise: QuantumNoiseError, gates: Union[str, List[str]], qubits: Union[int, List[int]] = None):
        """ Add noise which will affect target quantum gates with special qubits in the circuit.

        Args:
            noise (QuantumNoiseError): The noise error.
            gates (Union[str, List[str]]): The affected quantum gates.
            qubits (Union[int, List[int]], optional): The affected qubits, if None, same as add_noise_for_all_qubits.
                Defaults to None.
        """
        if qubits is None:
            self.add_noise_for_all_qubits(noise, gates)
            return

        # Input normalization
        qubits = self._qubits_normalize(qubits)
        gates = self._gates_normalize(noise, gates)

        # Add noise in the NoiseModel
        for gate in gates:
            self._error_by_gate[gate].append((noise, qubits))

    def add_noise_for_all_qubits(self, noise: QuantumNoiseError, gates: Union[str, List[str]]):
        """ Add noise which will affect all qubits in the circuit.

        Args:
            noise (QuantumNoiseError): The noise error
            gates (Union[str, List[str]]): The affected quantum gates.
        """
        # Input normalization
        gates = self._gates_normalize(noise, gates)

        # Add noise in the NoiseModel
        for gate in gates:
            self._error_by_gate[gate].append((noise, -1))

    def add_readout_error(self, noise: ReadoutError, qubits: Union[int, List[int]]):
        """ Add Readout error in the noise model.

        Example:
            1. Add 1-qubit Readout Error for all target qubits.
                noisemodel.add_readout_error(1-qubit ReadoutError, [0, 1, 2, etc])
            2. Add multi-qubits Readout Error for specified qubits (the number of qubits equal to ReadoutError.qubits).
                noisemodel.add_readout_error(multi-qubits ReadoutError, [1, 3, etc])

        Important:
            Supports to apply more than one Readout Error with the same target qubits, but do not support the
            combination of the Readout Error with the different qubits currently.

        Args:
            noise (ReadoutError): The Readout Error.
            qubits (Union[int, List[int]]): The target qubits for the Readout error
        """
        assert isinstance(noise, ReadoutError), TypeError("NoiseModel.addreadouterror", "ReadoutError", type(noise))
        qubits = self._qubits_normalize(qubits)

        if noise.qubits > 1:
            assert noise.qubits == len(qubits), NoiseApplyError(
                "For multi qubits Readout Error, the given qubits' number should equal to the noise's qubit number."
            )

        self._readout_errors.append((noise, qubits))

    def transpile(self, circuit: Circuit, accumulated_mode: bool = False):
        """ Apply all noise in the Noise Model to the given circuit, replaced related gate with the NoiseGate

        Args:
            circuit (Circuit): The given circuit.

        Returns:
            Circuit: The circuit with NoiseGates
        """
        qubit_num = circuit.width()
        noised_circuit = Circuit(qubit_num)
        for gate in circuit.gates:
            if isinstance(gate, BasicGate) and gate.type.name in self._basic_gates:
                gate_str = gate.type.name
                gate_args = gate.cargs + gate.targs
                noise_list = self._error_by_gate[gate_str]
                append_origin_gate = True
                for noise, qubits in noise_list:
                    if qubits == -1 or (set(qubits) & set(gate_args)) == set(gate_args):    # noise's qubit matched
                        if accumulated_mode or noise.type == NoiseChannel.damping:
                            noise_matrixs = noise.apply_to_gate(gate.matrix)
                            NoiseGate(noise_matrixs, len(gate_args)) & gate_args | noised_circuit
                            append_origin_gate = False
                        else:
                            prob = random()
                            noise_matrix = noise.prob_mapping_operator(prob)
                            Unitary(noise_matrix) | noised_circuit(gate_args)

                if append_origin_gate:
                    gate | noised_circuit
            else:
                gate | noised_circuit

        return noised_circuit

    def apply_readout_error(self, qureg):
        """ Apply readout error to target qubits.

        Args:
            qureg (Qureg): The circuits' qubits.
        """
        for readouterror, qubits in self._readout_errors:
            if readouterror.qubits == 1:
                for q in qubits:
                    if qureg[q].measured is None:
                        continue

                    truly_result = int(qureg[q])
                    qureg[q].measured = readouterror.apply_to_qubits(truly_result)
            else:
                all_qubits_measured = True
                for q in qubits:
                    if qureg[q].measured is None:
                        all_qubits_measured = False
                        break

                if not all_qubits_measured:
                    continue

                truly_result = int(qureg[qubits])
                noised_result = readouterror.apply_to_qubits(truly_result)
                for q in qubits[::-1]:
                    qureg[q].measured = noised_result & 1
                    noised_result >>= 1



import numpy as np


class SymbolicPhaseVariable:
    """
    Symbolic phase variable.
    """
    __slots__ = ['phase', 'label']

    def __init__(self, label):
        self.phase = None
        self.label = label

    def __rmul__(self, other):
        ret = SymbolicPhase()
        if isinstance(other, float) or isinstance(other, int):
            ret.var_dict[self.label] = [self, other]
        else:
            raise TypeError('Left operator must be number')
        return ret

    def __truediv__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return (1 / other) * self
        else:
            raise TypeError('Left operator must be number')

    def __neg__(self):
        return -1 * self


class SymbolicPhase:
    """
    Symbolic phase expression.
    """
    __slots__ = ['var_dict', 'const']

    def __init__(self):
        self.var_dict = {}
        self.const = 0

    def copy(self):
        """
        Get a copy of this SymbolicPhase.

        Returns:
            SymbolicPhase: a copy
        """
        ret = SymbolicPhase()
        for k, v in self.var_dict.items():
            ret.var_dict[k] = v.copy()
        ret.const = self.const
        return ret

    def evaluate(self):
        """
        Evaluate the value of this SymbolicPhase.
        If it contains any undetermined variable, it will return float('inf').

        Returns:
            float: value of this SymbolicPhase
        """
        ret = self.const
        for var, coef in self.var_dict.values():
            if np.isclose(coef, 0):
                continue
            if var.phase is None:
                return float('inf')
            ret += var.phase * coef
        return ret

    def __radd__(self, other):
        return self + other

    def __add__(self, other):
        ret = self.copy()
        if isinstance(other, SymbolicPhase):
            ret.const += other.const
            for key, pack in other.var_dict.items():
                var, val = pack
                if key not in ret.var_dict:
                    ret.var_dict[key] = [var, 0]
                ret.var_dict[key][1] += val
                if np.isclose(ret.var_dict[key][1], 0):
                    ret.var_dict.pop(key)
        elif isinstance(other, SymbolicPhaseVariable):
            if other.label not in ret.var_dict:
                ret.var_dict[other.label] = [other, 0]
            ret.var_dict[other.label][1] += 1
            if np.isclose(ret.var_dict[other.label][1], 0):
                ret.var_dict.pop(other)
        elif isinstance(other, int) or isinstance(other, float):
            ret.const += other
        else:
            raise TypeError("Invalid right operator.")
        return ret

    def __rmul__(self, other):
        ret = self.copy()
        if isinstance(other, float) or isinstance(other, int):
            ret.const *= other
            for each in ret.var_dict:
                ret.var_dict[each][1] *= other
        else:
            raise TypeError('Left operator must be number')
        return ret

    def __neg__(self):
        return -1 * self

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return -self + other

    def __float__(self):
        return self.evaluate()

    def __mod__(self, other):
        ret = self.copy()
        if isinstance(other, float) or isinstance(other, int):
            ret.const = np.mod(ret.const, other)
        else:
            raise TypeError('right operator mush be number')
        return ret
